import logging
import os
import re
from unittest import mock, skipUnless

from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs import settings
from pcs.cli.common.parse_args import InputModifiers
from pcs.lib.external import CommandRunner, is_service_enabled

# pylint: disable=invalid-name

testdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

runner = CommandRunner(
    mock.MagicMock(logging.Logger),
    MockLibraryReportProcessor(),
    os.environ
)


class ParametrizedTestMetaClass(type):
    """
    Example:
        class GeneralTest(TestCase):
            attr = None
            def _test_1(self):
                self.assertIn(self.attr, [1, 2])

            def _test_2(self):
                self.assertNotIn(self.attr, [0, 3, 4, 5])

        class Test1(GeneralTest, metaclass=ParametrizedTestMetaClass):
            attr = 1

        class Test2(GeneralTest, metaclass=ParametrizedTestMetaClass):
            attr = 2

        class Test3(GeneralTest, metaclass=ParametrizedTestMetaClass):
            # This should fail
            attr = 3
    """
    def __init__(cls, classname, bases, class_dict):
        for attr_name in dir(cls):
            attr = getattr(cls, attr_name)
            if (
                attr_name.startswith("_test")
                and
                hasattr(attr, "__call__")
            ):
                setattr(cls, attr_name[1:], attr)

        super().__init__(classname, bases, class_dict)


def dict_to_modifiers(options):
    def _convert_val(val):
        if val is True:
            return ""
        return val
    return InputModifiers(
        {
            f"--{opt}": _convert_val(val)
            for opt, val in options.items()
            if val is not False
        }
    )

def get_test_resource(name):
    """Return full path to a test resource file specified by name"""
    return os.path.join(testdir, "resources", name)

def read_test_resource(name):
    with open(get_test_resource(name)) as a_file:
        return a_file.read()

def cmp3(a, b):
    # python3 doesn't have the cmp function, this is an official workaround
    # https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
    return (a > b) - (a < b)

def compare_version(a, b):
    if a[0] == b[0]:
        if a[1] == b[1]:
            return cmp3(a[2], b[2])
        return cmp3(a[1], b[1])
    return cmp3(a[0], b[0])

def is_minimum_pacemaker_version(major, minor, rev):
    return is_version_sufficient(
        get_current_pacemaker_version(),
        (major, minor, rev)
    )

def get_current_pacemaker_version():
    output, dummy_stderr, dummy_retval = runner.run([
        os.path.join(settings.pacemaker_binaries, "crm_mon"),
        "--version",
    ])
    pacemaker_version = output.split("\n")[0]
    r = re.compile(r"Pacemaker (\d+)\.(\d+)\.(\d+)")
    m = r.match(pacemaker_version)
    major = int(m.group(1))
    minor = int(m.group(2))
    rev = int(m.group(3))
    return major, minor, rev

def is_version_sufficient(current_version, minimal_version):
    return compare_version(current_version, minimal_version) > -1

def format_version(version_tuple):
    return ".".join([str(x) for x in version_tuple])

def is_minimum_pacemaker_features(cmajor, cminor, crev):
    output, dummy_stderr, dummy_retval = runner.run([
        os.path.join(settings.pacemaker_binaries, "pacemakerd"),
        "--features",
    ])
    features_version = output.split("\n")[1]
    r = re.compile(r"Supporting v(\d+)\.(\d+)\.(\d+):")
    m = r.search(features_version)
    major = int(m.group(1))
    minor = int(m.group(2))
    rev = int(m.group(3))
    return compare_version((major, minor, rev), (cmajor, cminor, crev)) > -1

def skip_unless_pacemaker_version(version_tuple, feature):
    current_version = get_current_pacemaker_version()
    return skipUnless(
        is_version_sufficient(current_version, version_tuple),
        (
            "Pacemaker version is too old (current: {current_version},"
            " must be >= {minimal_version}) to test {feature}"
        ).format(
            current_version=format_version(current_version),
            minimal_version=format_version(version_tuple),
            feature=feature
        )
    )

skip_unless_crm_rule = skip_unless_pacemaker_version(
    (2, 0, 2),
    "listing of constraints that might be expired"
)

def skip_unless_pacemaker_features(version_tuple, feature):
    return skipUnless(
        is_minimum_pacemaker_features(*version_tuple),
        "Pacemaker must support feature set version {version} to test {feature}"
            .format(
                version=format_version(version_tuple),
                feature=feature
            )
    )

skip_unless_pacemaker_supports_bundle = skip_unless_pacemaker_features(
    (3, 1, 0),
    "bundle resources with promoted-max attribute"
)

def skip_if_service_enabled(service_name):
    return skipUnless(
        not is_service_enabled(runner, service_name),
        "Service {0} must be disabled".format(service_name),
    )

def skip_unless_root():
    return skipUnless(
        os.getuid() == 0,
        "Root user required"
    )

def create_patcher(target_prefix_or_module):
    """
    Return function for patching tests with preconfigured target prefix
    string|module target_prefix_or_module could be:
        * a prefix for patched names. Typicaly tested module:
            "pcs.lib.commands.booth"
        * a (imported) module: pcs.lib.cib
        Between prefix and target is "." (dot)
    """
    prefix = target_prefix_or_module
    if not isinstance(target_prefix_or_module, str):
        prefix = target_prefix_or_module.__name__

    def patch(target, *args, **kwargs):
        return mock.patch("{0}.{1}".format(prefix, target), *args, **kwargs)
    return patch

def outdent(text):
    line_list = text.splitlines()
    smallest_indentation = min([
        len(line) - len(line.lstrip(" "))
        for line in line_list if line
    ])
    return "\n".join([line[smallest_indentation:] for line in line_list])

def create_setup_patch_mixin(module_specification_or_patcher):
    """
    Configure and return SetupPatchMixin

    SetupPatchMixin add method 'setup_patch' to a test case.

    Method setup_patch takes name that should be patched in destination module
    (see module_specification_or_patcher). Method provide cleanup after test.
    It is expected to be used in 'setUp' method but should work inside test as
    well.

    string|callable module_specification_or_patcher can be
       * callable patcher created via create_patcher:
         create_patcher("pcs.lib.cib")
       * name of module: "pcs.lib.cib"
       * (imported) module: pcs.lib.cib
         Note that this must be not a callable (can be done via
         sys.modules[__name__] = something_callable. If is a callable use name
         of the module instead.
    """
    if callable(module_specification_or_patcher):
        patch_module = module_specification_or_patcher
    else:
        patch_module = create_patcher(module_specification_or_patcher)

    class SetupPatchMixin:
        def setup_patch(self, target_suffix, *args, **kwargs):
            patcher = patch_module(target_suffix, *args, **kwargs)
            self.addCleanup(patcher.stop)
            return patcher.start()
    return SetupPatchMixin



#!/usr/bin/python3
import sys
import os
import locale
import datetime

sys.path.insert(
    0,
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "pcs")
)
import settings


locale.setlocale(locale.LC_ALL, ("en_US", "UTF-8"))

# Get the current version, increment by 1, verify changes, git commit & tag
pcs_version_split = settings.pcs_version.split('.')
pcs_version_split[2] = str(int(pcs_version_split[2]) + 1)
new_version = ".".join(pcs_version_split)

print(os.system("sed -i 's/"+settings.pcs_version+"/"+new_version+"/' setup.py"))
print(os.system("sed -i 's/"+settings.pcs_version+"/"+new_version+"/' pcs/settings_default.py"))
print(os.system("sed -i 's/"+settings.pcs_version+"/"+new_version+"/' pcsd/bootstrap.rb"))
print(os.system("sed -i 's/\#\# \[Unreleased\]/\#\# ["+new_version+"] - "+datetime.date.today().strftime('%Y-%m-%d')+"/' CHANGELOG.md"))

def manpage_head(component, package="pcs"):
    return '.TH {component} "8" "{date}" "{package} {version}" "System Administration Utilities"'.format(
        component=component.upper(),
        date=datetime.date.today().strftime('%B %Y'),
        version=new_version,
        package=package,
    )
print(os.system("sed -i '1c " + manpage_head("pcs") + "' pcs/pcs.8"))
print(os.system("sed -i '1c " + manpage_head("pcsd") + "' pcsd/pcsd.8"))
print(os.system(
    "sed -i '1c {man_head}' pcs/snmp/pcs_snmp_agent.8".format(
        man_head=manpage_head("pcs_snmp_agent", package="pcs-snmp"),
    )
))

print(os.system("git diff"))
print("Look good? (y/n)")
choice = sys.stdin.read(1)
if choice != "y":
    print("Ok, exiting")
    sys.exit(0)

print(os.system("git commit -a -m 'Bumped to "+new_version+"'"))
print(os.system("git tag "+new_version))



from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal

from pcs.common import report_codes
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.booth import config_parser
from pcs.lib.booth.config_parser import ConfigItem


class BuildTest(TestCase):
    def test_build_file_content_from_parsed_structure(self):
        self.assertEqual(
            "\n".join([
                "authfile = /path/to/auth.file",
                "site = 1.1.1.1",
                "site = 2.2.2.2",
                "arbitrator = 3.3.3.3",
                'ticket = "TA"',
                'ticket = "TB"',
                "  timeout = 10",
                "", #newline at the end
            ]).encode("utf-8"),
            config_parser.Exporter.export([
                ConfigItem("authfile", "/path/to/auth.file"),
                ConfigItem("site", "1.1.1.1"),
                ConfigItem("site", "2.2.2.2"),
                ConfigItem("arbitrator", "3.3.3.3"),
                ConfigItem("ticket", "TA"),
                ConfigItem("ticket", "TB", [
                    ConfigItem("timeout", "10")
                ]),
            ])
        )


class OrganizeLinesTest(TestCase):
    def test_move_non_ticket_config_keys_above_tickets(self):
        self.assertEqual(
            [
                ConfigItem("site", "1.1.1.1"),
                ConfigItem('site', '2.2.2.2'),
                ConfigItem('arbitrator', '3.3.3.3'),
                ConfigItem("ticket", "TA"),
            ],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._organize_lines([
                ("site", "1.1.1.1"),
                ("ticket", "TA"),
                ('site', '2.2.2.2'),
                ('arbitrator', '3.3.3.3'),
            ])
        )

    def test_use_ticket_key_as_ticket_detail(self):
        # pylint: disable=invalid-name
        self.maxDiff = None
        self.assertEqual(
            [
                ConfigItem("site", "1.1.1.1"),
                ConfigItem('expire', '300'),
                ConfigItem('site', '2.2.2.2'),
                ConfigItem('arbitrator', '3.3.3.3'),
                ConfigItem("ticket", "TA", [
                    ConfigItem("timeout", "10"),
                    ConfigItem('--nonexistent', 'value'),
                    ConfigItem("expire", "300"),
                ]),
                ConfigItem("ticket", "TB", [
                    ConfigItem("timeout", "20"),
                    ConfigItem("renewal-freq", "40"),
                ]),
            ],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._organize_lines([
                ("site", "1.1.1.1"),
                ("expire", "300"), # out of ticket content is kept global
                ("ticket", "TA"),
                ("site", "2.2.2.2"), # move to global
                ("timeout", "10"),
                ("--nonexistent", "value"), # no global is kept under ticket
                ("expire", "300"),
                ("ticket", "TB"),
                ('arbitrator', '3.3.3.3'),
                ("timeout", "20"),
                ("renewal-freq", "40"),
            ])
        )


class ParseRawLinesTest(TestCase):
    def test_parse_simple_correct_lines(self):
        self.assertEqual(
            [
                ("site", "1.1.1.1"),
                ('site', '2.2.2.2'),
                ('arbitrator', '3.3.3.3'),
                ('syntactically_correct', 'nonsense'),
                ('line-with', 'hash#literal'),
            ],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines("\n".join([
                "site = 1.1.1.1",
                " site  =  2.2.2.2 ",
                "arbitrator=3.3.3.3",
                "syntactically_correct = nonsense",
                "line-with = hash#literal",
                "",
            ]))
        )

    def test_parse_lines_with_whole_line_comment(self):
        self.assertEqual(
            [("site", "1.1.1.1")],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines("\n".join([
                " # some comment",
                "site = 1.1.1.1",
            ]))
       )

    def test_skip_empty_lines(self):
        self.assertEqual(
            [("site", "1.1.1.1")],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines("\n".join([
                " ",
                "site = 1.1.1.1",
            ]))
       )

    def test_raises_when_unexpected_lines_appear(self):
        invalid_line_list = [
            "first invalid line",
            "second = 'invalid line' something else #comment",
            "third = 'invalid line 'something#'#",
        ]
        line_list = ["site = 1.1.1.1"] + invalid_line_list
        with self.assertRaises(config_parser.InvalidLines) as context_manager:
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines("\n".join(line_list))
        self.assertEqual(context_manager.exception.args[0], invalid_line_list)

    def test_parse_lines_finishing_with_comment(self):
        self.assertEqual(
            [("site", "1.1.1.1")],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines("\n".join([
                "site = '1.1.1.1' #comment",
            ]))
       )

class ParseTest(TestCase):
    def test_raises_when_invalid_lines_appear(self):
        invalid_line_list = [
            "first invalid line",
            "second = 'invalid line' something else #comment"
        ]
        line_list = ["site = 1.1.1.1"] + invalid_line_list

        with self.assertRaises(ParserErrorException) as cm:
            config_parser.Parser.parse("\n".join(line_list).encode("utf-8"))

        assert_report_item_list_equal(
            config_parser.Parser.exception_to_report_list(
                cm.exception, "does not matter", "path",
                None, False
            ),
            [
                fixture.error(
                    report_codes.BOOTH_CONFIG_UNEXPECTED_LINES,
                    line_list=invalid_line_list,
                    file_path="path",
                ),
            ]
        )
        assert_report_item_list_equal(
            config_parser.Parser.exception_to_report_list(
                cm.exception, "does not matter", "path",
                report_codes.FORCE_BOOTH_DESTROY, False
            ),
            [
                fixture.error(
                    report_codes.BOOTH_CONFIG_UNEXPECTED_LINES,
                    force_code=report_codes.FORCE_BOOTH_DESTROY,
                    line_list=invalid_line_list,
                    file_path="path",
                ),
            ]
        )
        assert_report_item_list_equal(
            config_parser.Parser.exception_to_report_list(
                cm.exception, "does not matter", "path",
                report_codes.FORCE_BOOTH_DESTROY, True
            ),
            [
                fixture.warn(
                    report_codes.BOOTH_CONFIG_UNEXPECTED_LINES,
                    line_list=invalid_line_list,
                    file_path="path",
                ),
            ]
        )

    def test_do_not_raises_when_no_invalid_liens_there(self):
        parsed = config_parser.Parser.parse("site = 1.1.1.1".encode("utf-8"))
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].key, "site")
        self.assertEqual(parsed[0].value, "1.1.1.1")
        self.assertEqual(parsed[0].details, [])



import shutil
import unittest

from pcs_test.tools.misc import (
    get_test_resource as rc,
    skip_unless_pacemaker_version,
    outdent,
    ParametrizedTestMetaClass,
)
from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.pcs_runner import PcsRunner


old_cib = rc("cib-empty-2.0.xml")
empty_cib = rc("cib-empty-2.5.xml")
temp_cib = rc("temp-cib.xml")

skip_unless_alerts_supported = skip_unless_pacemaker_version(
    (1, 1, 15),
    "alerts"
)

class PcsAlertTest(unittest.TestCase, AssertPcsMixin):
    def setUp(self):
        shutil.copy(empty_cib, temp_cib)
        self.pcs_runner = PcsRunner(temp_cib)


@skip_unless_alerts_supported
class AlertCibUpgradeTest(unittest.TestCase, AssertPcsMixin):
    def setUp(self):
        shutil.copy(old_cib, temp_cib)
        self.pcs_runner = PcsRunner(temp_cib)

    def test_cib_upgrade(self):
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 No alerts defined
"""
        )

        self.assert_pcs_success(
            "alert create path=test",
            "CIB has been upgraded to the latest schema version.\n"
        )

        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
"""
        )


@skip_unless_alerts_supported
class CreateAlertTest(PcsAlertTest):
    def test_create_multiple_without_id(self):
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 No alerts defined
"""
        )

        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success("alert create path=test2")
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
 Alert: alert-1 (path=test)
 Alert: alert-2 (path=test2)
"""
        )

    def test_create_multiple_with_id(self):
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 No alerts defined
"""
        )
        self.assert_pcs_success("alert create id=alert1 path=test")
        self.assert_pcs_success(
            "alert create id=alert2 description=desc path=test"
        )
        self.assert_pcs_success(
            "alert create description=desc2 path=test2 id=alert3"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert1 (path=test)
 Alert: alert2 (path=test)
  Description: desc
 Alert: alert3 (path=test2)
  Description: desc2
"""
        )

    def test_create_with_options(self):
        self.assert_pcs_success(
            "alert create id=alert1 description=desc path=test "
            "options opt2=val2 opt1=val1 meta m2=v2 m1=v1"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert1 (path=test)
  Description: desc
  Options: opt1=val1 opt2=val2
  Meta options: m1=v1 m2=v2
"""
        )

    def test_already_exists(self):
        self.assert_pcs_success("alert create id=alert1 path=test")
        self.assert_pcs_fail(
            "alert create id=alert1 path=test",
            "Error: 'alert1' already exists\n"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert1 (path=test)
"""
        )

    def test_path_is_required(self):
        self.assert_pcs_fail(
            "alert create id=alert1",
            "Error: required option 'path' is missing\n"
        )


@skip_unless_alerts_supported
class UpdateAlertTest(PcsAlertTest):
    def test_update_everything(self):
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 No alerts defined
"""
        )
        self.assert_pcs_success(
            "alert create id=alert1 description=desc path=test "
            "options opt1=val1 opt3=val3 meta m1=v1 m3=v3"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert1 (path=test)
  Description: desc
  Options: opt1=val1 opt3=val3
  Meta options: m1=v1 m3=v3
"""
        )
        self.assert_pcs_success(
            "alert update alert1 description=new_desc path=/new/path "
            "options opt1= opt2=test opt3=1 meta m1= m2=v m3=3"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert1 (path=/new/path)
  Description: new_desc
  Options: opt2=test opt3=1
  Meta options: m2=v m3=3
"""
        )

    def test_not_existing_alert(self):
        self.assert_pcs_fail(
            "alert update alert1", "Error: alert 'alert1' does not exist\n"
        )


class DeleteRemoveAlertTest(PcsAlertTest):
    command = None

    def _test_usage(self):
        self.assert_pcs_fail(
            f"alert {self.command}",
            stdout_start=f"\nUsage: pcs alert <command>\n    {self.command} <"
        )

    def _test_not_existing_alert(self):
        self.assert_pcs_fail(
            f"alert {self.command} alert1",
            "Error: alert 'alert1' does not exist\n"
        )

    def _test_one(self):
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 No alerts defined
                """
            )
        )

        self.assert_pcs_success("alert create path=test id=alert1")
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert1 (path=test)
                """
            )
        )
        self.assert_pcs_success(f"alert {self.command} alert1")
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 No alerts defined
                """
            )
        )

    def _test_multiple(self):
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 No alerts defined
                """
            )
        )

        self.assert_pcs_success("alert create path=test id=alert1")
        self.assert_pcs_success("alert create path=test id=alert2")
        self.assert_pcs_success("alert create path=test id=alert3")
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert1 (path=test)
                 Alert: alert2 (path=test)
                 Alert: alert3 (path=test)
                """
            )
        )
        self.assert_pcs_success(f"alert {self.command} alert1 alert3")
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert2 (path=test)
                """
            )
        )


@skip_unless_alerts_supported
class DeleteAlertTest(
    DeleteRemoveAlertTest,
    metaclass=ParametrizedTestMetaClass
):
    command = "delete"


@skip_unless_alerts_supported
class RemoveAlertTest(
    DeleteRemoveAlertTest,
    metaclass=ParametrizedTestMetaClass
):
    command = "remove"


@skip_unless_alerts_supported
class AddRecipientTest(PcsAlertTest):
    def test_success(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
"""
        )
        self.assert_pcs_success("alert recipient add alert value=rec_value")
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
"""
        )
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value2 id=my-recipient "
            "description=description options o2=2 o1=1 meta m2=v2 m1=v1"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
   Recipient: my-recipient (value=rec_value2)
    Description: description
    Options: o1=1 o2=2
    Meta options: m1=v1 m2=v2
"""
        )

    def test_already_exists(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec"
        )
        self.assert_pcs_fail(
            "alert recipient add alert value=value id=rec",
            "Error: 'rec' already exists\n"
        )
        self.assert_pcs_fail(
            "alert recipient add alert value=value id=alert",
            "Error: 'alert' already exists\n"
        )

    def test_same_value(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec"
        )
        self.assert_pcs_fail(
            "alert recipient add alert value=rec_value",
            "Error: Recipient 'rec_value' in alert 'alert' already exists, "
            "use --force to override\n"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: rec (value=rec_value)
"""
        )
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value --force",
            "Warning: Recipient 'rec_value' in alert 'alert' already exists\n"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: rec (value=rec_value)
   Recipient: alert-recipient (value=rec_value)
"""
        )

    def test_no_value(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_fail(
            "alert recipient add alert id=rec",
            "Error: required option 'value' is missing\n"
        )



@skip_unless_alerts_supported
class UpdateRecipientAlert(PcsAlertTest):
    def test_success(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value description=description "
            "options o1=1 o3=3 meta m1=v1 m3=v3"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
    Description: description
    Options: o1=1 o3=3
    Meta options: m1=v1 m3=v3
"""
        )
        self.assert_pcs_success(
            "alert recipient update alert-recipient value=new description=desc "
            "options o1= o2=v2 o3=3 meta m1= m2=2 m3=3"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=new)
    Description: desc
    Options: o2=v2 o3=3
    Meta options: m2=2 m3=3
"""
        )
        self.assert_pcs_success(
            "alert recipient update alert-recipient value=new"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=new)
    Description: desc
    Options: o2=v2 o3=3
    Meta options: m2=2 m3=3
"""
        )

    def test_value_exists(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success("alert recipient add alert value=rec_value")
        self.assert_pcs_success("alert recipient add alert value=value")
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
   Recipient: alert-recipient-1 (value=value)
"""
        )
        self.assert_pcs_fail(
            "alert recipient update alert-recipient value=value",
            "Error: Recipient 'value' in alert 'alert' already exists, "
            "use --force to override\n"
        )
        self.assert_pcs_success(
            "alert recipient update alert-recipient value=value --force",
            "Warning: Recipient 'value' in alert 'alert' already exists\n"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=value)
   Recipient: alert-recipient-1 (value=value)
"""
        )

    def test_value_same_as_previous(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success("alert recipient add alert value=rec_value")
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
"""
        )
        self.assert_pcs_success(
            "alert recipient update alert-recipient value=rec_value"
        )
        self.assert_pcs_success(
            "alert config",
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
"""
        )

    def test_no_recipient(self):
        self.assert_pcs_fail(
            "alert recipient update rec description=desc",
            "Error: recipient 'rec' does not exist\n"
        )

    def test_empty_value(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec"
        )
        self.assert_pcs_fail(
            "alert recipient update rec value=",
            "Error: Recipient value '' is not valid.\n"
        )


class DeleteRemoveRecipientTest(PcsAlertTest):
    command = None

    def _test_usage(self):
        self.assert_pcs_fail(
            f"alert recipient {self.command}",
            stdout_start=outdent(f"""
                Usage: pcs alert <command>
                    recipient {self.command} <""")
        )

    def _test_one(self):
        self.assert_pcs_success("alert create path=test")
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec"
        )
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert (path=test)
                  Recipients:
                   Recipient: rec (value=rec_value)
                """
            )
        )
        self.assert_pcs_success(f"alert recipient {self.command} rec")
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert (path=test)
                """
            )
        )

    def _test_multiple(self):
        self.assert_pcs_success("alert create path=test id=alert1")
        self.assert_pcs_success("alert create path=test id=alert2")
        self.assert_pcs_success(
            "alert recipient add alert1 value=rec_value1 id=rec1"
        )
        self.assert_pcs_success(
            "alert recipient add alert1 value=rec_value2 id=rec2"
        )
        self.assert_pcs_success(
            "alert recipient add alert2 value=rec_value3 id=rec3"
        )
        self.assert_pcs_success(
            "alert recipient add alert2 value=rec_value4 id=rec4"
        )
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert1 (path=test)
                  Recipients:
                   Recipient: rec1 (value=rec_value1)
                   Recipient: rec2 (value=rec_value2)
                 Alert: alert2 (path=test)
                  Recipients:
                   Recipient: rec3 (value=rec_value3)
                   Recipient: rec4 (value=rec_value4)
                """
            )
        )
        self.assert_pcs_success(
            f"alert recipient {self.command} rec1 rec2 rec4"
        )
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert1 (path=test)
                 Alert: alert2 (path=test)
                  Recipients:
                   Recipient: rec3 (value=rec_value3)
                """
            )
        )

    def _test_no_recipient(self):
        self.assert_pcs_success("alert create path=test id=alert1")
        self.assert_pcs_success(
            "alert recipient add alert1 value=rec_value1 id=rec1"
        )
        self.assert_pcs_fail(
            f"alert recipient {self.command} rec1 rec2 rec3", outdent("""\
                Error: recipient 'rec2' does not exist
                Error: recipient 'rec3' does not exist
                """
            )
        )
        self.assert_pcs_success(
            "alert config", outdent("""\
                Alerts:
                 Alert: alert1 (path=test)
                  Recipients:
                   Recipient: rec1 (value=rec_value1)
                """
            )
        )


class DeleteRecipientTest(
    DeleteRemoveRecipientTest,
    metaclass=ParametrizedTestMetaClass
):
    command = "delete"


class RemoveRecipientTest(
    DeleteRemoveRecipientTest,
    metaclass=ParametrizedTestMetaClass
):
    command = "remove"



def constraint_plain(constraint_info, with_id=False):
    """
    dict constraint_info see constraint in pcs/lib/exchange_formats.md
    bool with_id have to show id with options_dict
    """
    options_dict = constraint_info["options"]
    co_resource1 = options_dict.get("rsc", "")
    co_resource2 = options_dict.get("with-rsc", "")
    co_id = options_dict.get("id", "")
    co_score = options_dict.get("score", "")
    score_text = "(score:" + co_score + ")"
    console_option_list = [
        "(%s:%s)" % (option[0], option[1])
        for option in sorted(options_dict.items())
        if option[0] not in ("rsc", "with-rsc", "id", "score")
    ]
    if with_id:
        console_option_list.append("(id:%s)" % co_id)
    return " ".join(
        [co_resource1, "with", co_resource2, score_text] + console_option_list
    )



from pcs.common import report_codes as codes
from pcs.cli.common.console_report import (
    format_file_role,
    format_optional,
    format_plural,
)
from pcs.common.tools import format_list

def format_booth_default(value, template):
    return "" if value in ("booth", "", None) else template.format(value)

def booth_config_accepted_by_node(info):
    desc = ""
    if info["name_list"] and info["name_list"] not in [["booth"]]:
        desc = "{_s} {_list}".format(
            _s=("s" if len(info["name_list"]) > 1 else ""),
            _list=format_list(info["name_list"])
        )
    return "{_node_info}Booth config{_desc} saved".format(
        _node_info=format_optional(info["node"], "{0}: "),
        _desc=desc,
    )

#Each value (a callable taking report_item.info) returns a message.
#Force text will be appended if necessary.
#If it is necessary to put the force text inside the string then the callable
#must take the force_text parameter.
CODE_TO_MESSAGE_BUILDER_MAP = {
    codes.BOOTH_LACK_OF_SITES: lambda info:
        "lack of sites for booth configuration (need 2 at least): sites {0}"
        .format(", ".join(info["sites"]) if info["sites"] else "missing")
    ,

    codes.BOOTH_EVEN_PEERS_NUM: lambda info:
        "odd number of peers is required (entered {number} peers)"
        .format(**info)
    ,

    codes.BOOTH_ADDRESS_DUPLICATION: lambda info:
        "duplicate address for booth configuration: {0}"
        .format(", ".join(info["addresses"]))
    ,

    codes.BOOTH_CONFIG_UNEXPECTED_LINES: lambda info:
        "unexpected {_line_pl} in booth config{_file_path}:\n{_line_list}"
        .format(
            _file_path=format_optional(info["file_path"], " '{0}'"),
            _line_pl=format_plural(info["line_list"], "line"),
            _line_list="\n".join(info["line_list"]),
            **info
        )
    ,

    codes.BOOTH_INVALID_NAME: lambda info:
        "booth name '{name}' is not valid ({reason})"
        .format(**info)
    ,

    codes.BOOTH_TICKET_NAME_INVALID: lambda info:
        "booth ticket name '{0}' is not valid, use alphanumeric chars or dash"
        .format(info["ticket_name"])
    ,

    codes.BOOTH_TICKET_DUPLICATE: lambda info:
        "booth ticket name '{ticket_name}' already exists in configuration"
        .format(**info)
    ,

    codes.BOOTH_TICKET_DOES_NOT_EXIST: lambda info:
        "booth ticket name '{ticket_name}' does not exist"
        .format(**info)
    ,

    codes.BOOTH_ALREADY_IN_CIB: lambda info:
        "booth instance '{name}' is already created as cluster resource"
        .format(**info)
    ,

    codes.BOOTH_NOT_EXISTS_IN_CIB: lambda info:
        "booth instance '{name}' not found in cib"
        .format(**info)
    ,

    codes.BOOTH_CONFIG_IS_USED: lambda info:
        "booth instance '{0}' is used{1}".format(
            info["name"],
            " {0}".format(info["detail"]) if info["detail"] else "",
        )
    ,

    codes.BOOTH_MULTIPLE_TIMES_IN_CIB: lambda info:
        "found more than one booth instance '{name}' in cib"
        .format(**info)
    ,

    codes.BOOTH_CONFIG_DISTRIBUTION_STARTED: lambda info:
        "Sending booth configuration to cluster nodes..."
    ,

    codes.BOOTH_CONFIG_ACCEPTED_BY_NODE: booth_config_accepted_by_node,

    codes.BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR: lambda info:
        "Unable to save booth config{_desc} on node '{node}': {reason}".format(
            _desc=format_booth_default(info["name"], " '{0}'"),
            **info
        )
    ,

    codes.BOOTH_FETCHING_CONFIG_FROM_NODE: lambda info:
        "Fetching booth config{desc} from node '{node}'...".format(
            desc=format_booth_default(info["config"], " '{0}'"),
            **info
        )
    ,

    codes.BOOTH_DAEMON_STATUS_ERROR: lambda info:
        "unable to get status of booth daemon: {reason}".format(**info)
    ,

    codes.BOOTH_TICKET_STATUS_ERROR: "unable to get status of booth tickets",

    codes.BOOTH_PEERS_STATUS_ERROR: "unable to get status of booth peers",

    codes.BOOTH_CANNOT_DETERMINE_LOCAL_SITE_IP: lambda info:
        "cannot determine local site ip, please specify site parameter"
    ,

    codes.BOOTH_TICKET_OPERATION_FAILED: lambda info:
        (
            "unable to {operation} booth ticket '{ticket_name}'"
            " for site '{site_ip}', reason: {reason}"
        ).format(**info)

    ,

    codes.BOOTH_UNSUPPORTED_FILE_LOCATION: lambda info:
        (
            "{_file_role} '{file_path}' is outside of supported booth config "
            "directory '{expected_dir}', ignoring the file"
        ).format(
            _file_role=format_file_role(info["file_type_code"]),
            **info
        )
    ,
}



import errno
import os

# the import makes it look like RealFile is implemented here so we don't
# have to import RawFile from common and Ghost file from here in other
# places
# pylint: disable=unused-import
from pcs.common.file import(
    RawFile as RealFile,
    RawFileError,
    RawFileInterface,
)
from pcs.lib import reports

# TODO add logging (logger / debug reports ?)

def raw_file_error_report(error, force_code=None, is_forced_or_warning=False):
    """
    Translate a RawFileError instance to a report

    RawFileError error -- an exception to be translated
    string force_code -- is it a forcible error? by which code?
    bool is_forced_or_warning -- translate to a warning if True, error otherwise
    """
    return reports.get_problem_creator(force_code, is_forced_or_warning)(
        reports.file_io_error,
        error.metadata.file_type_code,
        error.action,
        error.reason,
        # do not report real file path if we were working with a ghost file
        file_path=(
            "" if isinstance(error, GhostFileError) else error.metadata.path
        ),
    )

def export_ghost_file(ghost_file):
    """
    Export GhostFile so it can be transfered to a client

    GhostFile ghost_file -- a ghost file instance
    """
    return {
        "content": ghost_file.content,
    }


class GhostFileError(RawFileError):
    pass


class GhostFile(RawFileInterface):
    def __init__(self, metadata, file_data=None):
        """
        FileMetadata metadata -- describes the file and provides its metadata
        bytes file_data -- data of the ghost file
        """
        super().__init__(metadata)
        self.__file_data = file_data

    @property
    def content(self):
        """
        Export the file content
        """
        return self.__file_data

    def exists(self):
        return self.__file_data is not None

    def read(self):
        if self.__file_data is None:
            raise GhostFileError(
                self.metadata,
                RawFileError.ACTION_READ,
                # get "no such file" message as defined and worded in the system
                os.strerror(errno.ENOENT)
            )
        return self.__file_data

    def write(self, file_data, can_overwrite=False):
        self.__file_data = file_data



from unittest import TestCase

from pcs.cli.common.completion import (
    _find_suggestions,
    has_applicable_environment,
    make_suggestions,
    _split_words,
)

tree = {
    "resource": {
        "op": {
            "add": {},
            "defaults": {},
            "remove": {},
        },
        "clone": {},
    },
    "cluster": {
        "auth": {},
        "cib": {},
    }
}

class SuggestionTest(TestCase):
    def test_suggest_nothing_when_cursor_on_first_word(self):
        self.assertEqual([], _find_suggestions(tree, ['pcs'], 0))
        self.assertEqual([], _find_suggestions(tree, ['pcs', 'resource'], 0))

    def test_suggest_nothing_when_cursor_possition_out_of_range(self):
        self.assertEqual([], _find_suggestions(tree, ['pcs', 'resource'], 3))

    def test_suggest_when_last_word_not_started(self):
        self.assertEqual(
            ["clone", "op"],
            _find_suggestions(tree, ['pcs', 'resource'], 2)
        )

    def test_suggest_when_last_word_started(self):
        self.assertEqual(
            ["clone"],
            _find_suggestions(tree, ['pcs', 'resource', 'c'], 2)
        )

    def test_suggest_when_cursor_on_word_amid(self):
        self.assertEqual(
            ["clone"],
            _find_suggestions(tree, ['pcs', 'resource', 'c', 'add'], 2)
        )

    def test_suggest_nothing_when_previously_typed_word_not_match(self):
        self.assertEqual(
            [],
            _find_suggestions(tree, ['pcs', 'invalid', 'c'], 2)
        )

class HasCompletionEnvironmentTest(TestCase):
    def test_returns_false_if_environment_inapplicable(self):
        inapplicable_environments = [
            {
                'COMP_CWORD': '1',
                'PCS_AUTO_COMPLETE': '1',
            },
            {
                'COMP_WORDS': 'pcs resource',
                'PCS_AUTO_COMPLETE': '1',
            },
            {
                'COMP_WORDS': 'pcs resource',
                'COMP_CWORD': '1',
            },
            {
                'COMP_WORDS': 'pcs resource',
                'COMP_CWORD': '1',
                'PCS_AUTO_COMPLETE': '0',
            },
            {
                'COMP_WORDS': 'pcs resource',
                'COMP_CWORD': '1a',
                'PCS_AUTO_COMPLETE': '1',
            },
            {
                'COMP_WORDS': 'pcs resource',
                'COMP_CWORD': '1',
                'PCS_AUTO_COMPLETE': '1',
            },
        ]
        for environment in inapplicable_environments:
            self.assertFalse(
                has_applicable_environment(environment),
                'environment evaluated as applicable (should not be): '
                +repr(environment)
            )

    def test_returns_true_if_environment_is_set(self):
        self.assertTrue(has_applicable_environment({
            "COMP_WORDS": "pcs resource",
            "COMP_CWORD": '1',
            "COMP_LENGTHS": "3 8",
            "PCS_AUTO_COMPLETE": "1",
        }))

class MakeSuggestionsEnvironment(TestCase):
    def test_raises_for_incomlete_environment(self):
        self.assertRaises(
            EnvironmentError,
            lambda: make_suggestions(
                {
                    'COMP_CWORD': '1',
                    'PCS_AUTO_COMPLETE': '1',
                },
                suggestion_tree=tree
            )
        )

    def test_suggest_on_correct_environment(self):
        self.assertEqual(
            "clone\nop",
            make_suggestions(
                {
                    "COMP_WORDS": "pcs resource",
                    "COMP_CWORD": "2",
                    "COMP_LENGTHS": "3 8",
                    "PCS_AUTO_COMPLETE": "1",
                },
                suggestion_tree=tree
            )
        )

class SplitWordsTest(TestCase):
    def test_return_word_list_on_compatible_words_and_lenght(self):
        self.assertEqual(
            ["pcs", "resource", "op", "a"],
            _split_words("pcs resource op a", ["3", "8", "2", "1"])
        )

    def test_refuse_when_no_int_in_lengths(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pcs resource op a", ["3", "8", "2", "A"])
        )

    def test_refuse_when_lengths_are_too_big(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pcs resource op a", ["3", "8", "2", "10"])
        )

    def test_refuse_when_separator_doesnot_match(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pc sresource op a", ["3", "8", "2", "1"])
        )

    def test_refuse_when_lengths_are_too_small(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pcs resource op a ", ["3", "8", "2", "1"])
        )



BOOTH_CONFIG = "BOOTH_CONFIG"
BOOTH_KEY = "BOOTH_KEY"
CIB = "CIB"
COROSYNC_AUTHKEY = "COROSYNC_AUTHKEY"
COROSYNC_CONF = "COROSYNC_CONF"
PACEMAKER_AUTHKEY = "PACEMAKER_AUTHKEY"
PCSD_ENVIRONMENT_CONFIG = "PCSD_ENVIRONMENT_CONFIG"
PCSD_SSL_CERT = "PCSD_SSL_CERT"
PCSD_SSL_KEY = "PCSD_SSL_KEY"
PCS_KNOWN_HOSTS = "PCS_KNOWN_HOSTS"
PCS_SETTINGS_CONF = "PCS_SETTINGS_CONF"



from pcs.common.file import (
    FileAlreadyExists,
    RawFileError,
    RawFileInterface,
)

CALL_TYPE_RAW_FILE_EXISTS = "CALL_TYPE_RAW_FILE_EXISTS"
CALL_TYPE_RAW_FILE_READ = "CALL_TYPE_RAW_FILE_READ"
CALL_TYPE_RAW_FILE_REMOVE = "CALL_TYPE_RAW_FILE_REMOVE"
CALL_TYPE_RAW_FILE_WRITE = "CALL_TYPE_RAW_FILE_WRITE"

class RawFileCall():
    def __init__(self, file_type_code, path):
        self.file_type_code = file_type_code
        self.path = path

    def _repr(self, method):
        return "<RawFile.{_method}> file_type_code={ftc} path={path}".format(
            _method=method,
            ftc=self.file_type_code,
            path=self.path
        )

class RawFileExistsCall(RawFileCall):
    type = CALL_TYPE_RAW_FILE_EXISTS

    def __init__(self, file_type_code, path, exists=True):
        super().__init__(file_type_code, path)
        self.exists = exists

    def __repr__(self):
        return super()._repr("exists")

class RawFileReadCall(RawFileCall):
    type = CALL_TYPE_RAW_FILE_READ

    def __init__(self, file_type_code, path, content=None, exception_msg=None):
        if content is None and exception_msg is None:
            raise RuntimeError(
                "Both 'content' and 'exception_msg' cannot be None"
            )
        if content is not None and exception_msg is not None:
            raise RuntimeError(
                "Both 'content' and 'exception_msg' cannot be set"
            )
        super().__init__(file_type_code, path)
        self.content = content
        self.exception_msg = exception_msg

    def __repr__(self):
        return super()._repr("read")

class RawFileWriteCall(RawFileCall):
    type = CALL_TYPE_RAW_FILE_WRITE

    def __init__(
        self, file_type_code, path, file_data, can_overwrite=False,
        already_exists=False, exception_msg=None, exception_action=None
    ):
        super().__init__(file_type_code, path)
        self.file_data = file_data
        self.can_overwrite = can_overwrite
        self.already_exists = already_exists
        self.exception_msg = exception_msg
        self.exception_action = exception_action

    def __repr__(self):
        return super()._repr("write")

class RawFileRemoveCall(RawFileCall):
    type = CALL_TYPE_RAW_FILE_REMOVE

    def __init__(
        self, file_type_code, path, fail_if_file_not_found=True,
        exception_msg=None, file_not_found_exception=False
    ):
        super().__init__(file_type_code, path)
        self.fail_if_file_not_found = fail_if_file_not_found
        self.exception_msg = exception_msg
        self.file_not_found_exception = file_not_found_exception

    def __repr__(self):
        return super()._repr("remove")

def _check_file_type_code_and_path(
    method, real_metadata, call_index, expected_call
):
    if real_metadata.file_type_code != expected_call.file_type_code:
        raise AssertionError(
            (
                "Trying to call RawFile.{method} (call no. {index}) "
                "with RawFile constructed for '{real_ftc}' but it is "
                "expected it to be constructed for '{expected_ftc}'"
            ).format(
                expected_ftc=expected_call.file_type_code,
                index=call_index,
                method=method,
                real_ftc=real_metadata.file_type_code,
            )
        )

    if real_metadata.path != expected_call.path:
        raise AssertionError(
            (
                "Trying to call RawFile.{method} (call no. {index}) "
                "for '{real_ftc}', real path is '{real_path}' but "
                "expected path is '{expected_path}'"
            ).format(
                expected_path=expected_call.path,
                index=call_index,
                method=method,
                real_ftc=real_metadata.file_type_code,
                real_path=real_metadata.path,
            )
        )

def get_raw_file_mock(call_queue):
    class RawFileMock(RawFileInterface):
        def exists(self):
            call_index, expected_call = call_queue.take(
                CALL_TYPE_RAW_FILE_EXISTS
            )
            _check_file_type_code_and_path(
                "exists", self.metadata, call_index, expected_call
            )
            return expected_call.exists

        def read(self):
            call_index, expected_call = call_queue.take(CALL_TYPE_RAW_FILE_READ)
            _check_file_type_code_and_path(
                "read", self.metadata, call_index, expected_call
            )
            if expected_call.exception_msg:
                raise RawFileError(
                    self.metadata,
                    RawFileError.ACTION_READ,
                    expected_call.exception_msg,
                )
            return expected_call.content

        def write(self, file_data, can_overwrite=False):
            call_index, expected_call = call_queue.take(
                CALL_TYPE_RAW_FILE_WRITE
            )
            _check_file_type_code_and_path(
                "write", self.metadata, call_index, expected_call
            )

            if file_data != expected_call.file_data:
                raise AssertionError(
                    (
                        "Trying to call RawFile.write (call no. {index}) "
                        "for '{real_ftc}', real written data is '{real_data}' "
                        "but expected data is '{expected_data}'"
                    ).format(
                        expected_data=expected_call.file_data,
                        index=call_index,
                        real_ftc=self.metadata.file_type_code,
                        real_data=file_data,
                    )
                )

            if can_overwrite != expected_call.can_overwrite:
                raise AssertionError(
                    (
                        "Trying to call RawFile.write (call no. {index}) "
                        "for '{real_ftc}', real can_overwrite is '{real_can}' "
                        "but expected can_overwrite is '{expected_can}'"
                    ).format(
                        expected_can=expected_call.can_overwrite,
                        index=call_index,
                        real_ftc=self.metadata.file_type_code,
                        real_can=can_overwrite
                    )
                )

            if expected_call.already_exists:
                raise FileAlreadyExists(self.metadata)
            if expected_call.exception_msg:
                raise RawFileError(
                    self.metadata,
                    (
                        expected_call.exception_action
                        if expected_call.exception_action
                        else RawFileError.ACTION_WRITE
                    ),
                    expected_call.exception_msg,
                )

        def remove(self, fail_if_file_not_found=True):
            call_index, expected_call = call_queue.take(
                CALL_TYPE_RAW_FILE_REMOVE
            )
            _check_file_type_code_and_path(
                "remove", self.metadata, call_index, expected_call
            )

            if fail_if_file_not_found != expected_call.fail_if_file_not_found:
                raise AssertionError(
                    (
                        "Trying to call RawFile.remove (call no. {index}) "
                        "for '{real_ftc}', real fail_if_file_not_found is "
                        "'{real_fail}' but expected fail_if_file_not_found is "
                        "'{expected_fail}'"
                    ).format(
                        expected_fail=expected_call.fail_if_file_not_found,
                        index=call_index,
                        real_ftc=self.metadata.file_type_code,
                        real_fail=fail_if_file_not_found,
                    )
                )

            exception_msg = expected_call.exception_msg
            if (
                expected_call.file_not_found_exception
                and
                fail_if_file_not_found
            ):
                exception_msg = "No such file or directory"
            if exception_msg:
                raise RawFileError(
                    self.metadata,
                    RawFileError.ACTION_REMOVE,
                    exception_msg,
                )

        def backup(self):
            # TODO implement
            raise NotImplementedError()

    return RawFileMock



import shutil
from unittest import TestCase

from pcs_test.tier0.cib_resource.common import get_cib_resources
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import  get_test_resource as rc
from pcs_test.tools.pcs_runner import PcsRunner


class OperationAdd(
    TestCase,
    get_assert_pcs_effect_mixin(get_cib_resources)
):
    temp_cib = rc("temp-cib.xml")
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.prepare_cib_file()
        self.pcs_runner = PcsRunner(self.temp_cib)

    def prepare_cib_file(self):
        with open(self.temp_cib, "w") as temp_cib_file:
            temp_cib_file.write(self.fixture_cib_cache())

    def fixture_cib_cache(self):
        if not hasattr(self.__class__, "cib_cache"):
            self.__class__.cib_cache = self.fixture_cib()
        return self.__class__.cib_cache

    def fixture_cib(self):
        shutil.copy(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib)
        self.assert_pcs_success(
            "resource create --no-default-ops R ocf:heartbeat:Dummy"
        )
        #add to cib:
        # <primitive class="ocf" id="R" provider="heartbeat" type="Dummy">
        #   <operations>
        #     <op id="R-monitor-interval-60s" interval="60s"
        #       name="monitor"
        #     />
        #   </operations>
        # </primitive>
        with open(self.temp_cib) as cib_file:
            cib_content = cib_file.read()

        #clean
        self.pcs_runner = None
        shutil.copy(self.empty_cib, self.temp_cib)

        return cib_content

    def test_base_add(self):
        self.assert_effect(
            "resource op add R start interval=20s",
            """<resources>
                <primitive class="ocf" id="R" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="R-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                        <op id="R-start-interval-20s" interval="20s"
                            name="start"
                        />
                    </operations>
                </primitive>
            </resources>"""
        )

    def test_add_with_OCF_CHECK_LEVEL(self):
        # pylint: disable=invalid-name
        self.assert_effect(
            "resource op add R start interval=20s OCF_CHECK_LEVEL=1"
                " description=test-description"
            ,
            """<resources>
                <primitive class="ocf" id="R" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="R-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                        <op description="test-description" name="start"
                            id="R-start-interval-20s" interval="20s"
                        >
                            <instance_attributes
                                id="params-R-start-interval-20s"
                            >
                                <nvpair
                                    id="R-start-interval-20s-OCF_CHECK_LEVEL-1"
                                    name="OCF_CHECK_LEVEL" value="1"
                                />
                            </instance_attributes>
                        </op>
                    </operations>
                </primitive>
            </resources>"""
        )

    def test_can_multiple_operation_add(self):
        self.assert_effect(
            "resource op add R start interval=20s",
            """<resources>
                <primitive class="ocf" id="R" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="R-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                        <op id="R-start-interval-20s" interval="20s"
                            name="start"
                        />
                    </operations>
                </primitive>
            </resources>"""
        )
        self.assert_effect(
            "resource op add R stop interval=30s",
            """<resources>
                <primitive class="ocf" id="R" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="R-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                        <op id="R-start-interval-20s" interval="20s"
                            name="start"
                        />
                        <op id="R-stop-interval-30s" interval="30s"
                            name="stop"
                        />
                    </operations>
                </primitive>
            </resources>"""
        )

    def test_id_specified(self):
        self.assert_effect(
            "resource op add R start timeout=30 id=abcd",
            """<resources>
                <primitive class="ocf" id="R" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="R-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                        <op id="abcd" interval="0s" name="start" timeout="30" />
                    </operations>
                </primitive>
            </resources>"""
        )

    def test_invalid_id(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource op add R start timeout=30 id=ab#cd",
            "Error: invalid operation id 'ab#cd', '#' is not a valid"
                " character for a operation id\n"
        )

    def test_duplicate_id(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource op add R start timeout=30 id=R",
            "Error: id 'R' is already in use, please specify another one\n"
        )

    def test_unknown_option(self):
        self.assert_pcs_fail(
            "resource op add R start timeout=30 requires=quorum",
            (
                "Error: requires is not a valid op option (use --force to "
                "override)\n"
            )
        )



from pcs_test.tools.command_env.calls import Queue, CallListBuilder
from pcs_test.tools.command_env.mock_runner import Runner as NewRunner


#TODO please remove this module when Runner  is not used. The only usage is
# in the module pcs_test.test_lib_commands_sbd currently.



class Runner:
    def __init__(self):
        self.set_runs([])

    def assert_everything_launched(self):
        if self.__call_queue.remaining:
            raise AssertionError(
                "There are remaining expected commands: \n    '{0}'".format(
                    "'\n    '".join([
                        call.command
                        for call in self.__call_queue.remaining
                    ])
                )
            )

    def set_runs(self, run_list):
        call_list_builder = CallListBuilder()
        for run in run_list:
            call_list_builder.place("", run)

        self.__call_queue = Queue(call_list_builder)
        self.run = NewRunner(self.__call_queue).run



from pcs.common import report_codes
from pcs.lib.errors import ReportItem, ReportItemSeverity


def booth_lack_of_sites(site_list):
    """
    Less than 2 booth sites entered. But it does not make sense.
    list site_list contains currently entered sites
    """
    return ReportItem.error(
        report_codes.BOOTH_LACK_OF_SITES,
        info={
            "sites": sorted(site_list),
        }
    )

def booth_even_peers_num(number):
    """
    Booth requires odd number of peers. But even number of peers was entered.
    integer number determines how many peers was entered
    """
    return ReportItem.error(
        report_codes.BOOTH_EVEN_PEERS_NUM,
        info={
            "number": number,
        }
    )

def booth_address_duplication(duplicate_addresses):
    """
    Address of each peer must unique. But address duplication appeared.
    set duplicate_addresses contains addreses entered multiple times
    """
    return ReportItem.error(
        report_codes.BOOTH_ADDRESS_DUPLICATION,
        info={
            "addresses": sorted(duplicate_addresses),
        }
    )

def booth_config_unexpected_lines(
    line_list, file_path=None,
    severity=ReportItemSeverity.ERROR, forceable=None,
):
    """
    Lines not conforming to expected config structure found in a booth config

    list line_list -- not valid lines
    string file_path -- path to the conf file if available
    """
    return ReportItem(
        report_codes.BOOTH_CONFIG_UNEXPECTED_LINES,
        severity,
        info={
            "line_list": line_list,
            "file_path": file_path,
        },
        forceable=forceable
    )

def booth_invalid_name(name, reason):
    """
    Booth instance name have rules. For example it cannot contain illegal
        characters like '/'. But some of rules was violated.
    string name is entered booth instance name
    """
    return ReportItem.error(
        report_codes.BOOTH_INVALID_NAME,
        info={
            "name": name,
            "reason": reason,
        }
    )

def booth_ticket_name_invalid(ticket_name):
    """
    Name of booth ticket may consists of alphanumeric characters or dash.
        Entered ticket name violating this rule.
    string ticket_name is entered booth ticket name
    """
    return ReportItem.error(
        report_codes.BOOTH_TICKET_NAME_INVALID,
        info={
            "ticket_name": ticket_name,
        }
    )

def booth_ticket_duplicate(ticket_name):
    """
    Each booth ticket name must be uniqe. But duplicate booth ticket name
        was entered.
    string ticket_name is entered booth ticket name
    """
    return ReportItem.error(
        report_codes.BOOTH_TICKET_DUPLICATE,
        info={
            "ticket_name": ticket_name,
        }
    )

def booth_ticket_does_not_exist(ticket_name):
    """
    Some operations (like ticket remove) expect the ticket name in booth
        configuration. But the ticket name not found in booth configuration.
    string ticket_name is entered booth ticket name
    """
    return ReportItem.error(
        report_codes.BOOTH_TICKET_DOES_NOT_EXIST,
        info={
            "ticket_name": ticket_name,
        }
    )

def booth_already_in_cib(name):
    """
    Each booth instance should be in a cib once maximally. Existence of booth
        instance in cib detected during creating new one.
    string name is booth instance name
    """
    return ReportItem.error(
        report_codes.BOOTH_ALREADY_IN_CIB,
        info={
            "name": name,
        }
    )

def booth_not_exists_in_cib(name):
    """
    Remove booth instance from cib required. But no such instance found in cib.
    string name is booth instance name
    """
    return ReportItem.error(
        report_codes.BOOTH_NOT_EXISTS_IN_CIB,
        info={
            "name": name,
        }
    )

def booth_config_is_used(name, detail=""):
    """
    Booth config use detected during destroy request.
    string name is booth instance name
    string detail provide more details (for example booth instance is used as
        cluster resource or is started/enabled under systemd)
    """
    return ReportItem.error(
        report_codes.BOOTH_CONFIG_IS_USED,
        info={
            "name": name,
            "detail": detail,
        }
    )

def booth_multiple_times_in_cib(
    name, severity=ReportItemSeverity.ERROR
):
    """
    Each booth instance should be in a cib once maximally. But multiple
        occurences detected. For example during remove booth instance from cib.
        Notify user about this fact is required. When operation is forced
        user should be notified about multiple occurences.
    string name is booth instance name
    ReportItemSeverity severit should be ERROR or WARNING (depends on context)
        is flag for next report processing
        Because of severity coupling with ReportItem is it specified here.
    """
    return ReportItem(
        report_codes.BOOTH_MULTIPLE_TIMES_IN_CIB,
        severity,
        info={
            "name": name,
        },
        forceable=report_codes.FORCE_BOOTH_REMOVE_FROM_CIB
            if severity == ReportItemSeverity.ERROR else None
    )


def booth_config_distribution_started():
    """
    booth configuration is about to be sent to nodes
    """
    return ReportItem.info(
        report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
    )


def booth_config_accepted_by_node(node=None, name_list=None):
    """
    Booth config has been saved on specified node.

    node -- name of node
    name_list -- list of names of booth instance
    """
    return ReportItem.info(
        report_codes.BOOTH_CONFIG_ACCEPTED_BY_NODE,
        info={
            "node": node,
            "name_list": sorted(name_list) if name_list else ""
        }
    )


def booth_config_distribution_node_error(node, reason, name=None):
    """
    Saving booth config failed on specified node.

    node -- node name
    reason -- reason of failure
    name -- name of booth instance
    """
    return ReportItem.error(
        report_codes.BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR,
        info={
            "node": node,
            "name": name,
            "reason": reason
        }
    )


def booth_fetching_config_from_node_started(node, config=None):
    """
    fetching of booth config from specified node started

    node -- node from which config is fetching
    config -- config name
    """
    return ReportItem.info(
        report_codes.BOOTH_FETCHING_CONFIG_FROM_NODE,
        info={
            "node": node,
            "config": config,
        }
    )


def booth_unsupported_file_location(file_path, expected_dir, file_type_code):
    """
    a booth file (config, authfile) is not in the expected dir, skipping it

    string file_path -- the actual path of the file
    string expected_dir -- where the file is supposed to be
    string file_type_code -- item from pcs.common.file_type_codes
    """
    return ReportItem.warning(
        report_codes.BOOTH_UNSUPPORTED_FILE_LOCATION,
        info={
            "file_path": file_path,
            "expected_dir": expected_dir,
            "file_type_code": file_type_code,
        }
    )


def booth_daemon_status_error(reason):
    """
    Unable to get status of booth daemon because of error.

    reason -- reason
    """
    return ReportItem.error(
        report_codes.BOOTH_DAEMON_STATUS_ERROR,
        info={"reason": reason}
    )


def booth_tickets_status_error(reason=None):
    """
    Unable to get status of booth tickets because of error.

    reason -- reason
    """
    return ReportItem.error(
        report_codes.BOOTH_TICKET_STATUS_ERROR,
        info={
            "reason": reason,
        }
    )


def booth_peers_status_error(reason=None):
    """
    Unable to get status of booth peers because of error.

    reason -- reason
    """
    return ReportItem.error(
        report_codes.BOOTH_PEERS_STATUS_ERROR,
        info={
            "reason": reason,
        }
    )

def booth_cannot_determine_local_site_ip():
    """
    Some booth operations are performed on specific site and requires to specify
        site ip. When site specification omitted pcs can try determine local ip.
        But determine local site ip failed.
    """
    return ReportItem.error(
        report_codes.BOOTH_CANNOT_DETERMINE_LOCAL_SITE_IP,
        info={}
    )

def booth_ticket_operation_failed(operation, reason, site_ip, ticket_name):
    """
    Pcs uses external booth tools for some ticket_name operations. For example
        grand and revoke. But the external command failed.
    string operatin determine what was intended perform with ticket_name
    string reason is taken from external booth command
    string site_ip specifiy what site had to run the command
    string ticket_name specify with which ticket had to run the command
    """
    return ReportItem.error(
        report_codes.BOOTH_TICKET_OPERATION_FAILED,
        info={
            "operation": operation,
            "reason": reason,
            "site_ip": site_ip,
            "ticket_name": ticket_name,
        }
    )



from unittest import TestCase

from pcs_test.tools.misc import outdent

class OutdentTest(TestCase):
    def test_returns_the_same_text_when_not_indented(self):
        text = "\n".join([
            "first line",
            "  second line",
            "    third line",
        ])
        self.assertEqual(text, outdent(text))

    def test_remove_the_smallest_indentation(self):
        self.assertEqual(
            "\n".join([
                "  first line",
                "second line",
                "  third line",
            ]),
            outdent("\n".join([
                "    first line",
                "  second line",
                "    third line",
            ]))
        )

    def test_very_ugly_indented_text(self):
        self.assertEqual(
            """\
Cluster Name: test99
  Options:
""",
            outdent("""\
                Cluster Name: test99
                  Options:
                """
            )
        )



from unittest import TestCase

from lxml import etree

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.xml import etree_to_str

from pcs.common import report_codes
from pcs.lib.commands import resource

# This class does not focusing on validation testing, there are validator tests
# for that in pcs_test.tier0.lib.cib.test_resource_relation
class GroupAdd(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_existing_group(self):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                </group>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
            </resources>
        """
        resources_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="R3" />
                    <primitive id="R1" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        (self.config
            .runner.cib.load(resources=resources_before)
            .env.push_cib(resources=resources_after)
        )
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R3", "R1"]
        )

    def test_new_group(self):
        resources_before = """
            <resources>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
            </resources>
        """
        resources_after = """
            <resources>
                <primitive id="R2" />
                <group id="G">
                    <primitive id="R3" />
                    <primitive id="R1" />
                </group>
            </resources>
        """
        (self.config
            .runner.cib.load(resources=resources_before)
            .env.push_cib(resources=resources_after)
        )
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R3", "R1"]
        )

    def _assert_with_adjacent(self, adjacent_id, after_adjacent):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RGX" />
                </group>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
            </resources>
        """
        resources_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="R3" />
                    <primitive id="RGX" />
                    <primitive id="R1" />
                    <primitive id="RG2" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        (self.config
            .runner.cib.load(resources=resources_before)
            .env.push_cib(resources=resources_after)
        )
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R3", "RGX", "R1"],
            adjacent_resource_id=adjacent_id,
            put_after_adjacent=after_adjacent
        )

    def test_after_adjacent(self):
        self._assert_with_adjacent("RG1", True)

    def test_before_adjacent(self):
        self._assert_with_adjacent("RG2", False)

    def test_remove_empty_containers(self):
        resources_before = """
            <resources>
                <group id="X">
                    <primitive id="RX1" />
                </group>
                <group id="Y">
                    <primitive id="RY1" />
                    <primitive id="RY2" />
                </group>
                <clone id="RC1-group-clone">
                    <group id="RC1-group">
                        <primitive id="RC1" />
                    </group>
                </clone>
                <master id="RM1-group-master">
                    <group id="RM1-group">
                        <primitive id="RM1" />
                    </group>
                </master>
            </resources>
        """
        resources_after = """
            <resources>
                <group id="Y">
                    <primitive id="RY1" />
                </group>
                <group id="G">
                    <primitive id="RX1" />
                    <primitive id="RY2" />
                    <primitive id="RC1" />
                    <primitive id="RM1" />
                </group>
            </resources>
        """
        (self.config
            .runner.cib.load(resources=resources_before)
            .env.push_cib(resources=resources_after)
        )
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["RX1", "RY2", "RC1", "RM1"]
        )

    def test_validation(self):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                </group>
                <primitive id="R1">
                    <meta_attributes id="R1-meta_attributes" />
                </primitive>
                <primitive id="R2">
                    <meta_attributes id="R2-meta_attributes" />
                </primitive>
                <clone id="RC1-clone">
                    <primitive id="RC1" />
                </clone>
            </resources>
        """
        (self.config
            .runner.cib.load(resources=resources_before)
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.group_add(
                self.env_assist.get_env(),
                "R1-meta_attributes",
                ["R2", "R4", "R3", "R2-meta_attributes", "RC1-clone", "RC1"]
            ),
            [
                fixture.error(
                    report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R1-meta_attributes",
                    expected_types=["group"],
                    current_type="meta_attributes",
                ),
                fixture.report_not_found("R4", context_type="resources"),
                fixture.report_not_found("R3", context_type="resources"),
                fixture.error(
                    report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R2-meta_attributes",
                    expected_types=[
                        "clone", "master", "group", "primitive", "bundle"
                    ],
                    current_type="meta_attributes",
                ),
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1-clone",
                    resource_type="clone",
                ),
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1",
                    resource_type="clone",
                ),
            ],
        )

    def test_validation_adjacent(self):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                </group>
                <group id="X">
                    <primitive id="RX1" />
                </group>
                <primitive id="R1" />
            </resources>
        """
        (self.config
            .runner.cib.load(resources=resources_before)
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.group_add(
                self.env_assist.get_env(),
                "G",
                ["R1"],
                adjacent_resource_id="RX1"
            ),
            [
                fixture.error(
                    report_codes
                        .CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP
                    ,
                    adjacent_resource_id="RX1",
                    group_id="G",
                ),
            ],
        )


class GroupAddWait(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        resources_before = """
            <resources>
                <primitive id="R1" />
                <primitive id="R2" />
            </resources>
        """
        resources_after = """
            <resources>
                <group id="G">
                    <primitive id="R1" />
                    <primitive id="R2" />
                </group>
            </resources>
        """
        self.timeout = 10
        (self.config
            .runner.pcmk.can_wait()
            .runner.cib.load(resources=resources_before)
            .env.push_cib(resources=resources_after, wait=self.timeout)
        )

    def test_group_running(self):
        resources_state = fixture.complete_state_resources(etree.fromstring("""
            <resources>
                <group id="G" number_resources="2">
                    <resource id="R1" role="Started" nodes_running_on="1">
                         <node name="node1" id="1" cached="false" />
                     </resource>
                    <resource id="R2" role="Started" nodes_running_on="1">
                         <node name="node1" id="1" cached="false" />
                     </resource>
                </group>
            </resources>
        """))
        self.config.runner.pcmk.load_state(
            resources=etree_to_str(resources_state)
        )
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R1", "R2"],
            wait=self.timeout,
        )
        self.env_assist.assert_reports([
            fixture.report_resource_running(
                "G",
                {"Started": ["node1"]}
            ),
        ])

    def test_group_not_running(self):
        resources_state = fixture.complete_state_resources(etree.fromstring("""
            <resources>
                <group id="G" number_resources="2">
                    <resource id="R1" role="Started" nodes_running_on="1">
                         <node name="node1" id="1" cached="false" />
                     </resource>
                    <resource id="R2" role="Stopped" nodes_running_on="0">
                     </resource>
                </group>
            </resources>
        """))
        self.config.runner.pcmk.load_state(
            resources=etree_to_str(resources_state)
        )
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R1", "R2"],
            wait=self.timeout,
        )
        self.env_assist.assert_reports([
            fixture.report_resource_not_running("G"),
        ])



from unittest import TestCase

from lxml import etree

from pcs.lib.cib import stonith


class IsStonithEnabled(TestCase):
    def test_not_set(self):
        crm_config = etree.fromstring("<crm_config />")
        self.assertTrue(stonith.is_stonith_enabled(crm_config))

    def test_set_to_enabled(self):
        crm_config = etree.fromstring("""
            <crm_config>
                <cluster_property_set>
                    <nvpair name="abc" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                </cluster_property_set>
            </crm_config>
        """)
        self.assertTrue(stonith.is_stonith_enabled(crm_config))

    def test_set_to_disabled(self):
        crm_config = etree.fromstring("""
            <crm_config>
                <cluster_property_set>
                    <nvpair name="abc" value="true" />
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """)
        self.assertFalse(stonith.is_stonith_enabled(crm_config))

    def test_multiple_values(self):
        crm_config = etree.fromstring("""
            <crm_config>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """)
        self.assertFalse(stonith.is_stonith_enabled(crm_config))

    def test_multiple_sections(self):
        crm_config = etree.fromstring("""
            <crm_config>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                </cluster_property_set>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """)
        self.assertFalse(stonith.is_stonith_enabled(crm_config))

class GetMisconfiguredResources(TestCase):
    def test_no_stonith(self):
        resources = etree.fromstring("""
            <resources>
                <primitive id="R" class="ocf" provider="pacemaker" type="Dummy">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
            </resources>
        """)
        self.assertEqual(
            stonith.get_misconfigured_resources(resources),
            ([], [], [])
        )

    def test_all_ok(self):
        resources = etree.fromstring("""
            <resources>
                <primitive id="S1" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="name" value="value" />
                    </instance_attributes>
                </primitive>
            </resources>
        """)
        self.assertEqual(
            stonith.get_misconfigured_resources(resources),
            (
                resources.findall("primitive[@id='S1']"),
                [],
                [],
            )
        )

    def test_issues(self):
        resources = etree.fromstring("""
            <resources>
                <primitive id="S1" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="name" value="value" />
                        <nvpair name="method" value="onoff" />
                    </instance_attributes>
                </primitive>
                <primitive id="S2" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                    </instance_attributes>
                </primitive>
                <primitive id="S3" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
                <primitive id="S4" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
            </resources>
        """)
        stonith1 = resources.find("primitive[@id='S1']")
        stonith2 = resources.find("primitive[@id='S2']")
        stonith3 = resources.find("primitive[@id='S3']")
        stonith4 = resources.find("primitive[@id='S4']")
        self.assertEqual(
            stonith.get_misconfigured_resources(resources),
            (
                [stonith1, stonith2, stonith3, stonith4],
                [stonith2, stonith4],
                [stonith3, stonith4],
            )
        )



from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
)

from pcs.common import report_codes
from pcs.lib.cib.resource import group
from pcs.lib.errors import ReportItemSeverity as severities


class IsGroup(TestCase):
    def test_is_group(self):
        self.assertTrue(group.is_group(etree.fromstring("<group/>")))
        self.assertFalse(group.is_group(etree.fromstring("<clone/>")))
        self.assertFalse(group.is_group(etree.fromstring("<master/>")))


@mock.patch("pcs.lib.cib.resource.group.find_element_by_tag_and_id")
class ProvideGroup(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(
            '<cib><resources><group id="g"/></resources></cib>'
        )
        self.group_element = self.cib.find('.//group')
        self.resources_section = self.cib.find('.//resources')

    def test_search_in_whole_tree(self, find_element_by_tag_and_id):
        def find_group(*args, **kwargs):
            # pylint: disable=unused-argument
            return self.group_element

        find_element_by_tag_and_id.side_effect = find_group

        self.assertEqual(
            self.group_element,
            group.provide_group(self.resources_section, "g")
        )

    def test_create_group_when_not_exists(self, find_element_by_tag_and_id):
        find_element_by_tag_and_id.return_value = None
        group_element = group.provide_group(self.resources_section, "g2")
        self.assertEqual('group', group_element.tag)
        self.assertEqual('g2', group_element.attrib["id"])

class PlaceResource(TestCase):
    def setUp(self):
        self.group_element = etree.fromstring("""
            <group id="g">
                <primitive id="a"/>
                <primitive id="b"/>
            </group>
        """)
        self.primitive_element = etree.Element("primitive", {"id": "c"})

    def assert_final_order(
        self, id_list=None, adjacent_resource_id=None, put_after_adjacent=False
    ):
        group.place_resource(
            self.group_element,
            self.primitive_element,
            adjacent_resource_id,
            put_after_adjacent
        )
        assert_xml_equal(
            etree.tostring(self.group_element).decode(),
            """
                <group id="g">
                    <primitive id="{0}"/>
                    <primitive id="{1}"/>
                    <primitive id="{2}"/>
                </group>
            """.format(*id_list)
        )

    def test_append_at_the_end_when_adjacent_is_not_specified(self):
        self.assert_final_order(["a", "b", "c"])

    def test_insert_before_adjacent(self):
        self.assert_final_order(["c", "a", "b"], "a")

    def test_insert_after_adjacent(self):
        self.assert_final_order(["a", "c", "b"], "a", put_after_adjacent=True)

    def test_insert_after_adjacent_which_is_last(self):
        self.assert_final_order(["a", "b", "c"], "b", put_after_adjacent=True)

    def test_refuse_to_put_next_to_the_same_resource_id(self):
        assert_raise_library_error(
            lambda: group.place_resource(
                self.group_element,
                self.primitive_element,
                adjacent_resource_id="c",
            ),
            (
                severities.ERROR,
                report_codes.CANNOT_GROUP_RESOURCE_NEXT_TO_ITSELF,
                {
                    "resource_id": "c",
                },
            ),
        )

    def test_raises_when_adjacent_resource_not_in_group(self):
        assert_raise_library_error(
            lambda: group.place_resource(
                self.group_element,
                self.primitive_element,
                adjacent_resource_id="r",
            ),
            (
                severities.ERROR,
                report_codes.ID_NOT_FOUND,
                {
                    "id": "r",
                    "expected_types": ["primitive"],
                    "context_type": "group",
                    "context_id": "g",
                },
                None
            ),
        )


class GetInnerResource(TestCase):
    def assert_inner_resource(self, resource_id, xml):
        self.assertEqual(
            resource_id,
            [
                element.attrib.get("id", "")
                for element in group.get_inner_resources(etree.fromstring(xml))
            ]
        )

    def test_one(self):
        self.assert_inner_resource(
            ["A"],
            """
                <group id="G">
                    <meta_attributes />
                    <primitive id="A" />
                    <meta_attributes />
                </group>
            """
        )

    def test_more(self):
        self.assert_inner_resource(
            ["A", "C", "B"],
            """
                <group id="G">
                    <meta_attributes />
                    <primitive id="A" />
                    <primitive id="C" />
                    <primitive id="B" />
                    <meta_attributes />
                </group>
            """
        )



import re

TRANSPORT_DEFAULT = "knet"
TRANSPORTS_KNET = ("knet", )
TRANSPORTS_UDP = ("udp", "udpu")
TRANSPORTS_ALL = TRANSPORTS_KNET + TRANSPORTS_UDP

LINKS_UDP_MIN = 1
LINKS_UDP_MAX = 1
LINKS_KNET_MIN = 1
LINKS_KNET_MAX = 8
LINKS_MAX = max(LINKS_KNET_MAX, LINKS_UDP_MAX)

NODE_OPTIONS = (
    ["name", "nodeid"]
    +
    [f"ring{i}_addr" for i in range(LINKS_MAX)]
)

IP_VERSION_4 = "ipv4"
IP_VERSION_6 = "ipv6"
IP_VERSION_46 = "ipv4-6"
IP_VERSION_64 = "ipv6-4"
IP_VERSION_VALUES = (
    IP_VERSION_4, IP_VERSION_6, IP_VERSION_46, IP_VERSION_64
)

# This must list ALL knet options even if they translate to themselves.
# The list is used to get a set of allowed options.
LINK_OPTIONS_KNET_TRANSLATION = (
    # (user facing name, corosync.conf name)
    ("linknumber", "linknumber"),
    ("link_priority", "knet_link_priority"),
    ("mcastport", "mcastport"),
    ("ping_interval", "knet_ping_interval"),
    ("ping_precision", "knet_ping_precision"),
    ("ping_timeout", "knet_ping_timeout"),
    ("pong_count", "knet_pong_count"),
    ("transport", "knet_transport"),
)
# This is what pcs expects to get from users. Actual names in corosync.conf are
# different.
LINK_OPTIONS_KNET_USER = tuple(sorted([
    pair[0] for pair in LINK_OPTIONS_KNET_TRANSLATION
]))
# This is what corsync works with.
LINK_OPTIONS_KNET_COROSYNC = tuple(sorted([
    pair[1] for pair in LINK_OPTIONS_KNET_TRANSLATION
]))

LINK_OPTIONS_UDP = (
    "bindnetaddr",
    "broadcast",
    "mcastaddr",
    "mcastport",
    "ttl",
)

QUORUM_OPTIONS = (
    "auto_tie_breaker",
    "last_man_standing",
    "last_man_standing_window",
    "wait_for_all",
)

QUORUM_OPTIONS_INCOMPATIBLE_WITH_QDEVICE = (
    "auto_tie_breaker",
    "last_man_standing",
    "last_man_standing_window",
)

OPTION_NAME_RE = re.compile(r"^[-_/a-zA-Z0-9]+$")
QUORUM_DEVICE_HEURISTICS_EXEC_NAME_RE = re.compile(r"^exec_[-_/a-zA-Z0-9]+$")



from unittest import TestCase

from pcs.cli.constraint_ticket import parse_args
from pcs.cli.common.errors import CmdLineInputError

class ParseAddTest(TestCase):
    def test_parse_add_args(self):
        self.assertEqual(
            parse_args.parse_add(
                ["T", "resource1", "ticket=T", "loss-policy=fence"]
            ),
            (
                "T",
                "resource1",
                "",
                {
                    "ticket": "T",
                    "loss-policy": "fence",
                }
            )
        )

    def test_parse_add_args_with_resource_role(self):
        self.assertEqual(
            parse_args.parse_add(
                ["T", "master", "resource1", "ticket=T", "loss-policy=fence"]
            ),
            (
                "T",
                "resource1",
                "master",
                {
                    "ticket": "T",
                    "loss-policy": "fence",
                }
            )
        )

    def test_raises_when_invalid_resource_specification(self):
        self.assertRaises(
            CmdLineInputError,
            lambda: parse_args.parse_add(
                ["T", "master", "resource1", "something_else"]
            )
        )

    def test_raises_when_ticket_and_resource_not_specified(self):
        self.assertRaises(
            CmdLineInputError,
            lambda: parse_args.parse_add(
                ["loss-policy=fence"]
            )
        )

    def test_raises_when_resource_not_specified(self):
        self.assertRaises(
            CmdLineInputError,
            lambda: parse_args.parse_add(
                ["T", "loss-policy=fence"]
            )
        )

class SeparateTailOptionCandidatesTest(TestCase):
    def test_separate_when_both_parts_there(self):
        self.assertEqual(
            (["a", "b"], ["c=d", "e=f"]),
            parse_args.separate_tail_option_candidates(["a", "b", "c=d", "e=f"])
        )

    def test_returns_empty_head_when_options_there_only(self):
        self.assertEqual(
            ([], ["c=d", "e=f"]),
            parse_args.separate_tail_option_candidates(["c=d", "e=f"])
        )

    def test_returns_empty_tail_when_no_options_there(self):
        self.assertEqual(
            (["a", "b"], []),
            parse_args.separate_tail_option_candidates(["a", "b"])
        )



import os.path
from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
    start_tag_error_text,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str, XmlManipulation

from pcs import settings
from pcs.common import report_codes
from pcs.common.tools import Version
import pcs.lib.pacemaker.live as lib
from pcs.lib.errors import ReportItemSeverity as Severity
from pcs.lib.external import CommandRunner

# pylint: disable=no-self-use

def get_runner(stdout="", stderr="", returncode=0, env_vars=None):
    runner = mock.MagicMock(spec_set=CommandRunner)
    runner.run.return_value = (stdout, stderr, returncode)
    runner.env_vars = env_vars if env_vars else {}
    return runner


class LibraryPacemakerTest(TestCase):
    @staticmethod
    def path(name):
        return os.path.join(settings.pacemaker_binaries, name)

    def crm_mon_cmd(self):
        return [self.path("crm_mon"), "--one-shot", "--as-xml", "--inactive"]


class GetClusterStatusXmlTest(LibraryPacemakerTest):
    def test_success(self):
        expected_stdout = "<xml />"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        real_xml = lib.get_cluster_status_xml(mock_runner)

        mock_runner.run.assert_called_once_with(self.crm_mon_cmd())
        self.assertEqual(expected_stdout, real_xml)

    def test_error(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.get_cluster_status_xml(mock_runner),
            (
                Severity.ERROR,
                report_codes.CRM_MON_ERROR,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )

        mock_runner.run.assert_called_once_with(self.crm_mon_cmd())

class GetClusterStatusText(TestCase):
    def setUp(self):
        self.mock_fencehistory_supported = mock.patch(
            "pcs.lib.pacemaker.live.is_fence_history_supported",
            return_value=True
        )
        self.mock_fencehistory_supported.start()
        self.expected_stdout = "cluster status"
        self.expected_stderr = ""
        self.expected_retval = 0

    def tearDown(self):
        self.mock_fencehistory_supported.stop()

    def get_runner(self, stdout=None, stderr=None, retval=None):
        return get_runner(
            self.expected_stdout if stdout is None else stdout,
            self.expected_stderr if stderr is None else stderr,
            self.expected_retval if retval is None else retval,
        )

    def test_success_minimal(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, False
        )

        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot", "--inactive"
        ])
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_verbose(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, True
        )

        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot", "--inactive", "--show-detail",
            "--show-node-attributes", "--failcounts", "--fence-history=3",
        ])
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_no_fence_history(self):
        self.mock_fencehistory_supported.stop()
        self.mock_fencehistory_supported = mock.patch(
            "pcs.lib.pacemaker.live.is_fence_history_supported",
            return_value=False
        )
        self.mock_fencehistory_supported.start()

        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, True
        )

        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot", "--inactive", "--show-detail",
            "--show-node-attributes", "--failcounts",
        ])
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_hide_inactive(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, True, False
        )

        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot"
        ])
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_hide_inactive_verbose(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, True, True
        )

        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot", "--show-detail",
            "--show-node-attributes", "--failcounts", "--fence-history=3",
        ])
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_error(self):
        mock_runner = self.get_runner("stdout", "stderr", 1)
        assert_raise_library_error(
            lambda: lib.get_cluster_status_text(mock_runner, False, False),
            (
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason="stderr\nstdout"
                )
            )
        )
        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot", "--inactive"
        ])

    def test_warnings(self):
        mock_runner = self.get_runner(
            stderr="msgA\nDEBUG: msgB\nmsgC\nDEBUG: msgd\n"
        )
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, False
        )

        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot", "--inactive"
        ])
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, ["msgA", "msgC"])

    def test_warnings_verbose(self):
        mock_runner = self.get_runner(
            stderr="msgA\nDEBUG: msgB\nmsgC\nDEBUG: msgd\n"
        )
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, True
        )

        mock_runner.run.assert_called_once_with([
            "/usr/sbin/crm_mon", "--one-shot", "--inactive", "--show-detail",
            "--show-node-attributes", "--failcounts", "--fence-history=3",
        ])
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(
            warnings,
            ["msgA", "DEBUG: msgB", "msgC", "DEBUG: msgd"]
        )

class GetCibXmlTest(LibraryPacemakerTest):
    def test_success(self):
        expected_stdout = "<xml />"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        real_xml = lib.get_cib_xml(mock_runner)

        mock_runner.run.assert_called_once_with(
            [self.path("cibadmin"), "--local", "--query"]
        )
        self.assertEqual(expected_stdout, real_xml)

    def test_error(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.get_cib_xml(mock_runner),
            (
                Severity.ERROR,
                report_codes.CIB_LOAD_ERROR,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )

        mock_runner.run.assert_called_once_with(
            [self.path("cibadmin"), "--local", "--query"]
        )

    def test_success_scope(self):
        expected_stdout = "<xml />"
        expected_stderr = ""
        expected_retval = 0
        scope = "test_scope"
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        real_xml = lib.get_cib_xml(mock_runner, scope)

        mock_runner.run.assert_called_once_with(
            [
                self.path("cibadmin"),
                "--local", "--query", "--scope={0}".format(scope)
            ]
        )
        self.assertEqual(expected_stdout, real_xml)

    def test_scope_error(self):
        expected_stdout = "some info"
        # yes, the numbers do not match, tested and verified with
        # pacemaker-2.0.0-1.fc29.1.x86_64
        expected_stderr = (
            "Call cib_query failed (-6): No such device or address"
        )
        expected_retval = 105
        scope = "test_scope"
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.get_cib_xml(mock_runner, scope=scope),
            (
                Severity.ERROR,
                report_codes.CIB_LOAD_ERROR_SCOPE_MISSING,
                {
                    "scope": scope,
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )

        mock_runner.run.assert_called_once_with(
            [
                self.path("cibadmin"),
                "--local", "--query", "--scope={0}".format(scope)
            ]
        )

class GetCibTest(LibraryPacemakerTest):
    def test_success(self):
        xml = "<xml />"
        assert_xml_equal(xml, str(XmlManipulation((lib.get_cib(xml)))))

    def test_invalid_xml(self):
        xml = "<invalid><xml />"
        assert_raise_library_error(
            lambda: lib.get_cib(xml),
            (
                Severity.ERROR,
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                {
                }
            )
        )

class Verify(LibraryPacemakerTest):
    def test_run_on_live_cib(self):
        runner = get_runner()
        self.assertEqual(
            lib.verify(runner),
            ("", "", 0, False)
        )
        runner.run.assert_called_once_with(
            [self.path("crm_verify"), "--live-check"],
        )

    def test_run_on_mocked_cib(self):
        fake_tmp_file = "/fake/tmp/file"
        runner = get_runner(env_vars={"CIB_file": fake_tmp_file})

        self.assertEqual(lib.verify(runner), ("", "", 0, False))
        runner.run.assert_called_once_with(
            [self.path("crm_verify"), "--xml-file", fake_tmp_file],
        )

    def test_run_verbose(self):
        runner = get_runner()
        self.assertEqual(
            lib.verify(runner, verbose=True),
            ("", "", 0, False)
        )
        runner.run.assert_called_once_with(
            [self.path("crm_verify"), "-V", "-V", "--live-check"],
        )

    def test_run_verbose_on_mocked_cib(self):
        fake_tmp_file = "/fake/tmp/file"
        runner = get_runner(env_vars={"CIB_file": fake_tmp_file})

        self.assertEqual(
            lib.verify(runner, verbose=True),
            ("", "", 0, False)
        )
        runner.run.assert_called_once_with(
            [self.path("crm_verify"), "-V", "-V", "--xml-file", fake_tmp_file],
        )

    @staticmethod
    def get_in_out_filtered_stderr():
        in_stderr = (
            (
                "Errors found during check: config not valid\n",
                "  -V may provide more details\n",
            ),
            (
                "Warnings found during check: config may not be valid\n",
                "  Use -V -V for more detail\n",
            ),
            (
                "some output\n",
                "another output\n",
                "-V -V -V more details...\n",
            ),
            (
                "some output\n",
                "before-V -V -V in the middle more detailafter\n",
                "another output\n",
            ),
        )
        out_stderr = []
        for input_lines in in_stderr:
            out_stderr.append([
                line for line in input_lines if "-V" not in line
            ])
        return zip(in_stderr, out_stderr)

    @staticmethod
    def get_in_out_unfiltered_data():
        in_out_data = (
            (
                "no verbose option in stderr\n",
            ),
            (
                "some output\n",
                "Options '-V -V' do not match\n",
                "because line missing 'more details'\n",
            ),
        )
        return zip(in_out_data, in_out_data)

    def subtest_filter_stderr_and_can_be_more_verbose(
        self, in_out_tuple_list, can_be_more_verbose, verbose=False,
    ):
        fake_tmp_file = "/fake/tmp/file"
        runner = get_runner(env_vars={"CIB_file": fake_tmp_file})
        for in_stderr, out_stderr in in_out_tuple_list:
            with self.subTest(in_stderr=in_stderr, out_stderr=out_stderr):
                runner = get_runner(
                    stderr="".join(in_stderr),
                    returncode=78,
                    env_vars={"CIB_file": fake_tmp_file},
                )
                self.assertEqual(
                    lib.verify(runner, verbose=verbose),
                    ("", "".join(out_stderr), 78, can_be_more_verbose),
                )
                args = [self.path("crm_verify")]
                if verbose:
                    args.extend(["-V", "-V"])
                args.extend(["--xml-file", fake_tmp_file])
                runner.run.assert_called_once_with(args)

    def test_error_can_be_more_verbose(self):
        self.subtest_filter_stderr_and_can_be_more_verbose(
            self.get_in_out_filtered_stderr(),
            True,
        )

    def test_error_cannot_be_more_verbose(self):
        self.subtest_filter_stderr_and_can_be_more_verbose(
            self.get_in_out_unfiltered_data(),
            False,
        )

    def test_error_cannot_be_more_verbose_in_verbose_mode(self):
        self.subtest_filter_stderr_and_can_be_more_verbose(
            (
                list(self.get_in_out_filtered_stderr())
                +
                list(self.get_in_out_unfiltered_data())
            ),
            False,
            verbose=True,
        )


class ReplaceCibConfigurationTest(LibraryPacemakerTest):
    def test_success(self):
        xml = "<xml/>"
        expected_stdout = "expected output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        lib.replace_cib_configuration(
            mock_runner,
            XmlManipulation.from_str(xml).tree
        )

        mock_runner.run.assert_called_once_with(
            [
                self.path("cibadmin"), "--replace", "--verbose", "--xml-pipe",
                "--scope", "configuration"
            ],
            stdin_string=xml
        )

    def test_error(self):
        xml = "<xml/>"
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.replace_cib_configuration(
                    mock_runner,
                    XmlManipulation.from_str(xml).tree
                )
            ,
            (
                Severity.ERROR,
                report_codes.CIB_PUSH_ERROR,
                {
                    "reason": expected_stderr,
                    "pushed_cib": expected_stdout,
                }
            )
        )

        mock_runner.run.assert_called_once_with(
            [
                self.path("cibadmin"), "--replace", "--verbose", "--xml-pipe",
                "--scope", "configuration"
            ],
            stdin_string=xml
        )

class UpgradeCibTest(TestCase):
    # pylint: disable=protected-access
    def test_success(self):
        mock_runner = get_runner("", "", 0)
        lib._upgrade_cib(mock_runner)
        mock_runner.run.assert_called_once_with(
            ["/usr/sbin/cibadmin", "--upgrade", "--force"]
        )

    def test_error(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )
        assert_raise_library_error(
            lambda: lib._upgrade_cib(mock_runner),
            (
                Severity.ERROR,
                report_codes.CIB_UPGRADE_FAILED,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )
        mock_runner.run.assert_called_once_with(
            ["/usr/sbin/cibadmin", "--upgrade", "--force"]
        )

@mock.patch("pcs.lib.pacemaker.live.get_cib_xml")
@mock.patch("pcs.lib.pacemaker.live._upgrade_cib")
class EnsureCibVersionTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)
        self.cib = etree.XML('<cib validate-with="pacemaker-2.3.4"/>')

    def test_same_version(self, mock_upgrade, mock_get_cib):
        self.assertTrue(
            lib.ensure_cib_version(
                self.mock_runner, self.cib, Version(2, 3, 4)
            ) is None
        )
        mock_upgrade.assert_not_called()
        mock_get_cib.assert_not_called()

    def test_higher_version(self, mock_upgrade, mock_get_cib):
        self.assertTrue(
            lib.ensure_cib_version(
                self.mock_runner, self.cib, Version(2, 3, 3)
            ) is None
        )
        mock_upgrade.assert_not_called()
        mock_get_cib.assert_not_called()

    def test_upgraded_same_version(self, mock_upgrade, mock_get_cib):
        upgraded_cib = '<cib validate-with="pacemaker-2.3.5"/>'
        mock_get_cib.return_value = upgraded_cib
        assert_xml_equal(
            upgraded_cib,
            etree.tostring(
                lib.ensure_cib_version(
                    self.mock_runner, self.cib, Version(2, 3, 5)
                )
            ).decode()
        )
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)

    def test_upgraded_higher_version(self, mock_upgrade, mock_get_cib):
        upgraded_cib = '<cib validate-with="pacemaker-2.3.6"/>'
        mock_get_cib.return_value = upgraded_cib
        assert_xml_equal(
            upgraded_cib,
            etree.tostring(
                lib.ensure_cib_version(
                    self.mock_runner, self.cib, Version(2, 3, 5)
                )
            ).decode()
        )
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)

    def test_upgraded_lower_version(self, mock_upgrade, mock_get_cib):
        mock_get_cib.return_value = etree.tostring(self.cib).decode()
        assert_raise_library_error(
            lambda: lib.ensure_cib_version(
                self.mock_runner, self.cib, Version(2, 3, 5)
            ),
            (
                Severity.ERROR,
                report_codes.CIB_UPGRADE_FAILED_TO_MINIMAL_REQUIRED_VERSION,
                {
                    "required_version": "2.3.5",
                    "current_version": "2.3.4"
                }
            )
        )
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)

    def test_cib_parse_error(self, mock_upgrade, mock_get_cib):
        mock_get_cib.return_value = "not xml"
        assert_raise_library_error(
            lambda: lib.ensure_cib_version(
                self.mock_runner, self.cib, Version(2, 3, 5)
            ),
            (
                Severity.ERROR,
                report_codes.CIB_UPGRADE_FAILED,
                {
                    "reason":
                        start_tag_error_text(),
                }
            )
        )
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)


@mock.patch("pcs.lib.pacemaker.live.write_tmpfile")
class SimulateCibXml(LibraryPacemakerTest):
    def test_success(self, mock_write_tmpfile):
        tmpfile_new_cib = mock.MagicMock()
        tmpfile_new_cib.name = rc("new_cib.tmp")
        tmpfile_new_cib.read.return_value = "new cib data"
        tmpfile_transitions = mock.MagicMock()
        tmpfile_transitions.name = rc("transitions.tmp")
        tmpfile_transitions.read.return_value = "transitions data"
        mock_write_tmpfile.side_effect = [tmpfile_new_cib, tmpfile_transitions]

        expected_stdout = "simulate output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        result = lib.simulate_cib_xml(mock_runner, "<cib />")
        self.assertEqual(result[0], expected_stdout)
        self.assertEqual(result[1], "transitions data")
        self.assertEqual(result[2], "new cib data")

        mock_runner.run.assert_called_once_with(
            [
                self.path("crm_simulate"),
                "--simulate",
                "--save-output", tmpfile_new_cib.name,
                "--save-graph", tmpfile_transitions.name,
                "--xml-pipe",
            ],
            stdin_string="<cib />"
        )

    def test_error_creating_cib(self, mock_write_tmpfile):
        mock_write_tmpfile.side_effect = OSError(1, "some error")
        mock_runner = get_runner()
        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason="some error",
                cib="<cib />",
            ),
        )
        mock_runner.run.assert_not_called()

    def test_error_creating_transitions(self, mock_write_tmpfile):
        tmpfile_new_cib = mock.MagicMock()
        mock_write_tmpfile.side_effect = [
            tmpfile_new_cib,
            OSError(1, "some error")
        ]
        mock_runner = get_runner()
        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason="some error",
                cib="<cib />",
            ),
        )
        mock_runner.run.assert_not_called()

    def test_error_running_simulate(self, mock_write_tmpfile):
        tmpfile_new_cib = mock.MagicMock()
        tmpfile_new_cib.name = rc("new_cib.tmp")
        tmpfile_transitions = mock.MagicMock()
        tmpfile_transitions.name = rc("transitions.tmp")
        mock_write_tmpfile.side_effect = [tmpfile_new_cib, tmpfile_transitions]

        expected_stdout = "some stdout"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason="some error",
                cib="<cib />",
            ),
        )

    def test_error_reading_cib(self, mock_write_tmpfile):
        tmpfile_new_cib = mock.MagicMock()
        tmpfile_new_cib.name = rc("new_cib.tmp")
        tmpfile_new_cib.read.side_effect = OSError(1, "some error")
        tmpfile_transitions = mock.MagicMock()
        tmpfile_transitions.name = rc("transitions.tmp")
        mock_write_tmpfile.side_effect = [tmpfile_new_cib, tmpfile_transitions]

        expected_stdout = "simulate output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason="some error",
                cib="<cib />",
            ),
        )

    def test_error_reading_transitions(self, mock_write_tmpfile):
        tmpfile_new_cib = mock.MagicMock()
        tmpfile_new_cib.name = rc("new_cib.tmp")
        tmpfile_new_cib.read.return_value = "new cib data"
        tmpfile_transitions = mock.MagicMock()
        tmpfile_transitions.name = rc("transitions.tmp")
        tmpfile_transitions.read.side_effect = OSError(1, "some error")
        mock_write_tmpfile.side_effect = [tmpfile_new_cib, tmpfile_transitions]

        expected_stdout = "simulate output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason="some error",
                cib="<cib />",
            ),
        )


@mock.patch("pcs.lib.pacemaker.live.simulate_cib_xml")
class SimulateCib(TestCase):
    def setUp(self):
        self.runner = "mock runner"
        self.cib_xml = "<cib/>"
        self.cib = etree.fromstring(self.cib_xml)
        self.simulate_output = "  some output  "
        self.transitions = "<transitions/>"
        self.new_cib = "<new-cib/>"

    def test_success(self, mock_simulate):
        mock_simulate.return_value = (
            self.simulate_output, self.transitions, self.new_cib
        )
        result = lib.simulate_cib(self.runner, self.cib)
        self.assertEqual(result[0], "some output")
        self.assertEqual(etree_to_str(result[1]), self.transitions)
        self.assertEqual(etree_to_str(result[2]), self.new_cib)
        mock_simulate.assert_called_once_with(self.runner, self.cib_xml)

    def test_invalid_cib(self, mock_simulate):
        mock_simulate.return_value = (
            self.simulate_output, "bad transitions", self.new_cib
        )
        assert_raise_library_error(
            lambda: lib.simulate_cib(self.runner, self.cib),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=(
                    "Start tag expected, '<' not found, line 1, column 1 "
                    "(<string>, line 1)"
                ),
                cib=self.cib_xml,
            ),
        )

    def test_invalid_transitions(self, mock_simulate):
        mock_simulate.return_value = (
            self.simulate_output, self.transitions, "bad new cib"
        )
        assert_raise_library_error(
            lambda: lib.simulate_cib(self.runner, self.cib),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=(
                    "Start tag expected, '<' not found, line 1, column 1 "
                    "(<string>, line 1)"
                ),
                cib=self.cib_xml,
            ),
        )


class GetLocalNodeStatusTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_offline(self):
        (self.config
            .runner.pcmk.load_state(
                stderr="error: Could not connect to cluster (is it running?)",
                returncode=102
            )
        )

        env = self.env_assist.get_env()
        real_status = lib.get_local_node_status(env.cmd_runner())
        self.assertEqual(dict(offline=True), real_status)

    def test_invalid_status(self):
        (self.config
            .runner.pcmk.load_state(stdout="invalid xml")
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_local_node_status(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.BAD_CLUSTER_STATE_FORMAT,
                    force_code=None
                )
            ],
            expected_in_processor=False
        )

    def test_success(self):
        (self.config
            .runner.pcmk.load_state(nodes=[
                fixture.state_node(i, f"name_{i}") for i in range(1, 4)
            ])
            .runner.pcmk.local_node_name(node_name="name_2")
        )

        env = self.env_assist.get_env()
        real_status = lib.get_local_node_status(env.cmd_runner())
        self.assertEqual(
            dict(offline=False, **fixture.state_node("2", "name_2")),
            real_status
        )

    def test_node_not_in_status(self):
        (self.config
            .runner.pcmk.load_state(nodes=[
                fixture.state_node(i, f"name_{i}") for i in range(1, 4)
            ])
            .runner.pcmk.local_node_name(node_name="name_X")
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_local_node_status(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    force_code=None,
                    node="name_X",
                    searched_types=[]
                )
            ],
            expected_in_processor=False
        )

    def test_error_getting_node_name(self):
        (self.config
            .runner.pcmk.load_state(nodes=[
                fixture.state_node(i, f"name_{i}") for i in range(1, 4)
            ])
            .runner.pcmk.local_node_name(
                stdout="some info", stderr="some error", returncode=1
            )
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_local_node_status(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.PACEMAKER_LOCAL_NODE_NAME_NOT_FOUND,
                    force_code=None,
                    reason="some error\nsome info"
                )
            ],
            expected_in_processor=False
        )


class RemoveNode(LibraryPacemakerTest):
    def test_success(self):
        mock_runner = get_runner("", "", 0)
        lib.remove_node(
            mock_runner,
            "NODE_NAME"
        )
        mock_runner.run.assert_called_once_with([
            self.path("crm_node"),
            "--force",
            "--remove",
            "NODE_NAME",
        ])

    def test_error(self):
        expected_stderr = "expected stderr"
        mock_runner = get_runner("", expected_stderr, 1)
        assert_raise_library_error(
            lambda: lib.remove_node(mock_runner, "NODE_NAME"),
            (
                Severity.ERROR,
                report_codes.NODE_REMOVE_IN_PACEMAKER_FAILED,
                {
                    "node": None,
                    "node_list_to_remove": ["NODE_NAME"],
                    "reason": expected_stderr,
                }
            )
        )


class ResourceCleanupTest(TestCase):
    def setUp(self):
        self.stdout = "expected output"
        self.stderr = "expected stderr"
        self.resource = "my_resource"
        self.node = "my_node"
        self.env_assist, self.config = get_env_tools(test_case=self)

    def assert_output(self, real_output):
        self.assertEqual(
            self.stdout + "\n" + self.stderr,
            real_output
        )

    def test_basic(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout,
            stderr=self.stderr
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(env.cmd_runner())
        self.assert_output(real_output)

    def test_resource(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout,
            stderr=self.stderr,
            resource=self.resource
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(
            env.cmd_runner(), resource=self.resource
        )
        self.assert_output(real_output)

    def test_node(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout,
            stderr=self.stderr,
            node=self.node
        )

        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(
            env.cmd_runner(), node=self.node
        )
        self.assert_output(real_output)

    def test_all_options(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout,
            stderr=self.stderr,
            resource=self.resource,
            node=self.node
        )

        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(
            env.cmd_runner(), resource=self.resource, node=self.node
        )
        self.assert_output(real_output)

    def test_error_cleanup(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout,
            stderr=self.stderr,
            returncode=1
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.resource_cleanup(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.RESOURCE_CLEANUP_ERROR,
                    force_code=None,
                    reason=(self.stderr + "\n" + self.stdout)
                )
            ],
            expected_in_processor=False
        )


class ResourceRefreshTest(LibraryPacemakerTest):
    def fixture_status_xml(self, nodes, resources):
        xml_man = XmlManipulation.from_file(rc("crm_mon.minimal.xml"))
        doc = xml_man.tree.getroottree()
        doc.find("/summary/nodes_configured").set("number", str(nodes))
        doc.find("/summary/resources_configured").set("number", str(resources))
        return str(XmlManipulation(doc))

    def test_basic(self):
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        mock_runner = mock.MagicMock(spec_set=CommandRunner)
        call_list = [
            mock.call(self.crm_mon_cmd()),
            mock.call([self.path("crm_resource"), "--refresh"]),
        ]
        return_value_list = [
            (self.fixture_status_xml(1, 1), "", 0),
            (expected_stdout, expected_stderr, 0),
        ]
        mock_runner.run.side_effect = return_value_list

        real_output = lib.resource_refresh(mock_runner)

        self.assertEqual(len(return_value_list), len(call_list))
        self.assertEqual(len(return_value_list), mock_runner.run.call_count)
        mock_runner.run.assert_has_calls(call_list)
        self.assertEqual(
            expected_stdout + "\n" + expected_stderr,
            real_output
        )

    def test_threshold_exceeded(self):
        mock_runner = get_runner(
            self.fixture_status_xml(1000, 1000),
            "",
            0
        )

        assert_raise_library_error(
            lambda: lib.resource_refresh(mock_runner),
            (
                Severity.ERROR,
                report_codes.RESOURCE_REFRESH_TOO_TIME_CONSUMING,
                {"threshold": 100},
                report_codes.FORCE_LOAD_THRESHOLD
            )
        )

        mock_runner.run.assert_called_once_with(self.crm_mon_cmd())

    def test_threshold_exceeded_forced(self):
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        mock_runner = get_runner(expected_stdout, expected_stderr, 0)

        real_output = lib.resource_refresh(mock_runner, force=True)

        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "--refresh"]
        )
        self.assertEqual(
            expected_stdout + "\n" + expected_stderr,
            real_output
        )

    def test_resource(self):
        resource = "test_resource"
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        mock_runner = get_runner(expected_stdout, expected_stderr, 0)

        real_output = lib.resource_refresh(mock_runner, resource=resource)

        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "--refresh", "--resource", resource]
        )
        self.assertEqual(
            expected_stdout + "\n" + expected_stderr,
            real_output
        )

    def test_node(self):
        node = "test_node"
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        mock_runner = get_runner(expected_stdout, expected_stderr, 0)

        real_output = lib.resource_refresh(mock_runner, node=node)

        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "--refresh", "--node", node]
        )
        self.assertEqual(
            expected_stdout + "\n" + expected_stderr,
            real_output
        )

    def test_full(self):
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        mock_runner = mock.MagicMock(spec_set=CommandRunner)
        call_list = [
            mock.call(self.crm_mon_cmd()),
            mock.call([self.path("crm_resource"), "--refresh", "--force"]),
        ]
        return_value_list = [
            (self.fixture_status_xml(1, 1), "", 0),
            (expected_stdout, expected_stderr, 0),
        ]
        mock_runner.run.side_effect = return_value_list

        real_output = lib.resource_refresh(mock_runner, full=True)

        self.assertEqual(len(return_value_list), len(call_list))
        self.assertEqual(len(return_value_list), mock_runner.run.call_count)
        mock_runner.run.assert_has_calls(call_list)
        self.assertEqual(
            expected_stdout + "\n" + expected_stderr,
            real_output
        )

    def test_all_options(self):
        node = "test_node"
        resource = "test_resource"
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        mock_runner = get_runner(expected_stdout, expected_stderr, 0)

        real_output = lib.resource_refresh(
            mock_runner, resource=resource, node=node, full=True
        )

        mock_runner.run.assert_called_once_with(
            [
                self.path("crm_resource"),
                "--refresh", "--resource", resource, "--node", node, "--force"
            ]
        )
        self.assertEqual(
            expected_stdout + "\n" + expected_stderr,
            real_output
        )

    def test_error_state(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.resource_refresh(mock_runner),
            (
                Severity.ERROR,
                report_codes.CRM_MON_ERROR,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )

        mock_runner.run.assert_called_once_with(self.crm_mon_cmd())

    def test_error_refresh(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = mock.MagicMock(spec_set=CommandRunner)
        call_list = [
            mock.call(self.crm_mon_cmd()),
            mock.call([self.path("crm_resource"), "--refresh"]),
        ]
        return_value_list = [
            (self.fixture_status_xml(1, 1), "", 0),
            (expected_stdout, expected_stderr, expected_retval),
        ]
        mock_runner.run.side_effect = return_value_list

        assert_raise_library_error(
            lambda: lib.resource_refresh(mock_runner),
            (
                Severity.ERROR,
                report_codes.RESOURCE_REFRESH_ERROR,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )

        self.assertEqual(len(return_value_list), len(call_list))
        self.assertEqual(len(return_value_list), mock_runner.run.call_count)
        mock_runner.run.assert_has_calls(call_list)


class ResourcesWaitingTest(LibraryPacemakerTest):
    def test_has_support(self):
        expected_stdout = ""
        expected_stderr = "something --wait something else"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        self.assertTrue(
            lib.has_wait_for_idle_support(mock_runner)
        )
        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "-?"]
        )

    def test_has_support_stdout(self):
        expected_stdout = "something --wait something else"
        expected_stderr = ""
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        self.assertTrue(
            lib.has_wait_for_idle_support(mock_runner)
        )
        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "-?"]
        )

    def test_doesnt_have_support(self):
        expected_stdout = "something something else"
        expected_stderr = "something something else"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        self.assertFalse(
            lib.has_wait_for_idle_support(mock_runner)
        )
        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "-?"]
        )

    @mock.patch(
        "pcs.lib.pacemaker.live.has_wait_for_idle_support",
        autospec=True
    )
    def test_ensure_support_success(self, mock_obj):
        mock_obj.return_value = True
        self.assertEqual(None, lib.ensure_wait_for_idle_support(mock.Mock()))

    @mock.patch(
        "pcs.lib.pacemaker.live.has_wait_for_idle_support",
        autospec=True
    )
    def test_ensure_support_error(self, mock_obj):
        mock_obj.return_value = False
        assert_raise_library_error(
            lambda: lib.ensure_wait_for_idle_support(mock.Mock()),
            (
                Severity.ERROR,
                report_codes.WAIT_FOR_IDLE_NOT_SUPPORTED,
                {}
            )
        )

    def test_wait_success(self):
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        self.assertEqual(None, lib.wait_for_idle(mock_runner))

        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "--wait"]
        )

    def test_wait_timeout_success(self):
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        expected_retval = 0
        timeout = 10
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        self.assertEqual(None, lib.wait_for_idle(mock_runner, timeout))

        mock_runner.run.assert_called_once_with(
            [
                self.path("crm_resource"),
                "--wait", "--timeout={0}".format(timeout)
            ]
        )

    def test_wait_error(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.wait_for_idle(mock_runner),
            (
                Severity.ERROR,
                report_codes.WAIT_FOR_IDLE_ERROR,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )

        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "--wait"]
        )

    def test_wait_error_timeout(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 124
        mock_runner = get_runner(
            expected_stdout,
            expected_stderr,
            expected_retval
        )

        assert_raise_library_error(
            lambda: lib.wait_for_idle(mock_runner),
            (
                Severity.ERROR,
                report_codes.WAIT_FOR_IDLE_TIMED_OUT,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                }
            )
        )

        mock_runner.run.assert_called_once_with(
            [self.path("crm_resource"), "--wait"]
        )



from unittest import TestCase

from pcs_test.tier0.lib.commands.resource.bundle_common import(
    FixturesMixin,
    SetUpMixin,
    UpgradeMixin,
    ParametrizedContainerMixin,
    NetworkMixin,
    PortMapMixin,
    StorageMapMixin,
    MetaMixin,
    AllOptionsMixin,
    WaitMixin,
)

from pcs.common import report_codes
from pcs.lib.commands import resource
from pcs.lib.errors import ReportItemSeverity as severities

class CreateCommandMixin:
    container_type = None
    bundle_id = "B1"
    image = "pcs:test"

    def bundle_create(self, bundle_id=None, container_type=None, **params):
        if "container_options" not in params:
            params["container_options"] = {"image": self.image}

        resource.bundle_create(
            self.env_assist.get_env(),
            bundle_id=bundle_id or self.bundle_id,
            container_type=container_type or self.container_type,
            **params
        )

    def run_bundle_cmd(self, *args, **kwargs):
        self.bundle_create(*args, **kwargs)


class MinimalCreate(CreateCommandMixin, FixturesMixin, SetUpMixin, TestCase):
    container_type = "docker"

    def test_success(self):
        self.config.env.push_cib(resources=self.fixture_resources_bundle_simple)
        self.bundle_create()

    def test_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda:
            self.bundle_create(bundle_id="B#1", container_type="nonsense"),
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_ID,
                    {
                        "invalid_character": "#",
                        "id": "B#1",
                        "id_description": "bundle name",
                        "is_first_char": False,
                    },
                    None
                ),
                (
                    severities.ERROR,
                    report_codes.INVALID_OPTION_VALUE,
                    {
                        "option_name": "container type",
                        "option_value": "nonsense",
                        "allowed_values": {"docker", "podman", "rkt"},
                        "cannot_be_empty": False,
                        "forbidden_characters": None,
                    },
                    None
                ),
            ]
        )

class CreateParametrizedContainerMixin(
    CreateCommandMixin, ParametrizedContainerMixin, UpgradeMixin
):
    pass

class CreateDocker(CreateParametrizedContainerMixin, TestCase):
    container_type = "docker"
    old_version_cib_filename = "cib-empty-2.0.xml"

class CreatePodman(CreateParametrizedContainerMixin, TestCase):
    container_type = "podman"
    old_version_cib_filename = "cib-empty-3.1.xml"


class CreateRkt(CreateParametrizedContainerMixin, TestCase):
    container_type = "rkt"
    old_version_cib_filename = "cib-empty-2.9.xml"

class CreateWithNetwork(CreateCommandMixin, NetworkMixin, TestCase):
    container_type = "docker"

class CreateWithPortMap(CreateCommandMixin, PortMapMixin, TestCase):
    container_type = "docker"

class CreateWithStorageMap(CreateCommandMixin, StorageMapMixin, TestCase):
    container_type = "docker"

class CreateWithMeta(CreateCommandMixin, MetaMixin, TestCase):
    container_type = "docker"

class CreateWithAllOptions(CreateCommandMixin, AllOptionsMixin, TestCase):
    container_type = "docker"

class Wait(CreateCommandMixin, WaitMixin, TestCase):
    container_type = "docker"



import io
import socket
from unittest import mock

from pcs_test.tools.assertions import assert_report_item_list_equal

import pcs.common.pcs_pycurl as pycurl
from pcs.cli.common.reports import LibraryReportProcessorToConsole
from pcs.lib.errors import LibraryError, ReportItemSeverity

def get_getaddrinfo_mock(resolvable_addr_list):
    def socket_getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
        # pylint: disable=redefined-builtin, unused-argument
        if host not in resolvable_addr_list:
            raise socket.gaierror(1, "")
    return socket_getaddrinfo

def patch_getaddrinfo(test_case, addr_list):
    """
    class MyTest(TestCase):
        def setUp(self):
            self.resolvable_hosts = patch_getaddrinfo(self, [])
        def test_something(self):
            self.resolvable_hosts.extend(["node1", "node2"])
    """
    patcher = mock.patch("socket.getaddrinfo", get_getaddrinfo_mock(addr_list))
    patcher.start()
    test_case.addCleanup(patcher.stop)
    return addr_list


class MockLibraryReportProcessor(LibraryReportProcessorToConsole):
    def __init__(self, debug=False, raise_on_errors=True):
        super(MockLibraryReportProcessor, self).__init__(debug)
        self.raise_on_errors = raise_on_errors
        self.direct_sent_items = []

    @property
    def report_item_list(self):
        return self.items

    def report_list(self, report_list):
        self.direct_sent_items.extend(report_list)
        return self._send(report_list)

    def send(self):
        errors = self._send(self.items, print_errors=False)
        if errors and self.raise_on_errors:
            raise LibraryError(*errors)

    def assert_reports(self, expected_report_info_list, hint=""):
        assert_report_item_list_equal(
            self.report_item_list + self.direct_sent_items,
            expected_report_info_list,
            hint=hint
        )

    def _send(self, report_item_list, print_errors=True):
        errors = []
        for report_item in report_item_list:
            if report_item.severity == ReportItemSeverity.ERROR:
                errors.append(report_item)
        return errors



class MockCurl:
    # pylint: disable=too-many-instance-attributes
    def __init__(
            self, info=None, output=b"", debug_output_list=None, exception=None,
            error=None, request=None
    ):
        # we don't need exception anymore, because we don't use perform on
        # easy hanlers. but for now it has to stay as it is because it is sill
        # used from old communicator tests
        self._opts = {}
        self._info = info if info else {}
        self._output = output
        self._debug_output_list = debug_output_list if debug_output_list else []
        self._error = error
        self._exception = exception
        self.request_obj = request

    @property
    def opts(self):
        return self._opts

    @property
    def error(self):
        return self._error

    def reset(self):
        self._opts = {}

    def setopt(self, opt, val):
        if isinstance(val, list):
            # in tests we use set operations (e.g. assertLessEqual) which
            # require hashable values
            val = tuple(val)
        if val is None:
            self.unsetopt(opt)
        else:
            self._opts[opt] = val

    def unsetopt(self, opt):
        try:
            del self._opts[opt]
        except KeyError:
            pass

    def getinfo(self, opt):
        try:
            return self._info[opt]
        except KeyError:
            AssertionError("info '#{0}' not defined".format(opt))

    def perform(self):
        if self._error:
            return
        if self._exception:
            #pylint: disable=raising-bad-type
            raise self._exception
        if pycurl.WRITEFUNCTION in self._opts:
            self._opts[pycurl.WRITEFUNCTION](self._output)
        if pycurl.DEBUGFUNCTION in self._opts:
            for msg_type, msg in self._debug_output_list:
                self._opts[pycurl.DEBUGFUNCTION](msg_type, msg)


class MockCurlSimple:
    def __init__(self, info=None, output=b"", debug_output=b"", request=None):
        self.output_buffer = io.BytesIO()
        self.output_buffer.write(
            output if isinstance(output, bytes) else output.encode("utf-8")
        )
        self.debug_buffer = io.BytesIO()
        self.debug_buffer.write(
            debug_output if isinstance(debug_output, bytes)
            else debug_output.encode("utf-8")
        )
        self.request_obj = request
        self._info = info if info else {}

    def getinfo(self, opt):
        try:
            return self._info[opt]
        except KeyError:
            AssertionError("info '#{0}' not defined".format(opt))


class MockCurlMulti:
    def __init__(self, number_of_performed_list):
        self._number_of_performed_list = number_of_performed_list
        self._opts = {}
        self._handle_list = []
        self._proccessed_list = []

    @property
    def opts(self):
        return self._opts

    def setopt(self, opt, val):
        self._opts[opt] = val

    def add_handle(self, handle):
        if not isinstance(handle, MockCurl):
            raise AssertionError("Only MockCurl objects are allowed")
        if handle in self._handle_list:
            # same error as real CurlMulti object
            raise pycurl.error("curl object already on this multi-stack")
        self._handle_list.append(handle)

    def remove_handle(self, handle):
        if handle not in self._handle_list:
            # same error as real CurlMulti object
            raise pycurl.error("curl object not on this multi-stack")
        self._handle_list.remove(handle)

    def assert_no_handle_left(self):
        if self._handle_list:
            raise AssertionError(
                "{0} handle(s) left to process".format(len(self._handle_list))
            )

    def select(self, timeout=1):
        # pylint: disable=no-self-use, unused-argument
        return 0

    def perform(self):
        # pylint: disable=no-self-use
        return (0, 0)

    def timeout(self):
        # pylint: disable=no-self-use
        return 0

    def info_read(self):
        ok_list = []
        err_list = []
        if not self._number_of_performed_list:
            raise AssertionError("unexpected info_read call")
        number_to_perform = self._number_of_performed_list.pop(0)
        if number_to_perform > len(self._handle_list):
            raise AssertionError("expecting more handles than prepared")
        for handle in self._handle_list[:number_to_perform]:
            try:
                handle.perform()
                if handle.error:
                    err_list.append((handle, handle.error[0], handle.error[1]))
                else:
                    ok_list.append(handle)
            except pycurl.error as e:
                # pylint: disable=unbalanced-tuple-unpacking
                errno, msg = e.args
                err_list.append((handle, errno, msg))
            self._proccessed_list.append(handle)
        return (0, ok_list, err_list)



from pcs import settings, utils
from pcs.cli.common.errors import CmdLineInputError


def local_auth_cmd(lib, argv, modifiers):
    """
    Options:
      * -u - username
      * -p - password
      * --request-timeout - timeout for HTTP requests
    """
    del lib
    modifiers.ensure_only_supported("-u", "-p", "--request-timeout")
    if len(argv) > 1:
        raise CmdLineInputError()
    port = argv[0] if argv else settings.pcsd_default_port
    username, password = utils.get_user_and_pass()
    utils.auth_hosts(
        {
            "localhost": {
                "username": username,
                "password": password,
                "dest_list": [{"addr": "localhost", "port": port}]
            }
        }
    )



from unittest import TestCase
import logging

from pcs_test.tools.misc import create_setup_patch_mixin

from pcs.daemon import auth

USER = "user"
PASSWORD = "password"

# Don't write errors to test output.
logging.getLogger("pcsd.daemon").setLevel(logging.CRITICAL)

class AuthorizeUserSync(TestCase, create_setup_patch_mixin(auth)):
    def setUp(self):
        self.authenticate_by_pam = self.setup_patch("authenticate_by_pam")
        self.get_user_groups_sync = self.setup_patch("get_user_groups_sync")

    def test_success_authorization(self):
        self.authenticate_by_pam.return_value = True
        self.get_user_groups_sync.return_value = [auth.HA_ADM_GROUP]

        user_auth_info = auth.authorize_user_sync(USER, PASSWORD)
        self.assertEqual(user_auth_info.name, USER)
        self.assertTrue(user_auth_info.is_authorized)

    def test_failed_pam_authorization(self):
        self.authenticate_by_pam.return_value = False

        user_auth_info = auth.authorize_user_sync(USER, PASSWORD)
        self.assertEqual(user_auth_info.name, USER)
        self.assertFalse(user_auth_info.is_authorized)

    def test_failed_on_groups(self):
        self.authenticate_by_pam.return_value = True
        self.get_user_groups_sync.return_value = []

        user_auth_info = auth.authorize_user_sync(USER, PASSWORD)
        self.assertEqual(user_auth_info.name, USER)
        self.assertFalse(user_auth_info.is_authorized)
